# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Curve"

mashCurveId = om.MTypeId(0x0011BE27)

class mashCurve( omMPx.MPxNode ):
    
    outputArray = om.MObject()
    inputCurve = om.MObject()
    
    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_CurveTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Curve" "Curve";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Curve Node" -collapse 0;
        
        editorTemplate  -label "Time Step" -addControl "timeStep";
        editorTemplate  -label "Animation (slide)" -addControl "timeSlide";
        editorTemplate -addSeparator;
        
        editorTemplate  -label "Strength" -addControl "Envelope";
        editorTemplate  -label "Random Strength" -addControl "randEnvelope";
        editorTemplate  -label "Step Strength" -addControl "StepEnvelope";
        editorTemplate -addSeparator;
        editorTemplate  -label "Calculate Rotation" -addControl "calculateRotation";
        editorTemplate  -label "Stop At Curve End" -addControl "stopAtEnd";
        editorTemplate -addSeparator;
        
        editorTemplate -suppress "inIterations";
        
        editorTemplate -beginLayout "Inputs" -collapse 0;
        
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "inputCurve" "Input Curve";
        
        editorTemplate -callCustom "AEcurveButton" "AEcurveButtonEdit" "";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        global proc AEcurveButton ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Set Curve"  -c ("curveButtonCMDS " + $nodeName[0] + " 1")
        setCurveMASHButton;
        

        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        global proc AEcurveButtonEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("curveButtonCMDS " + $nodeName[0] + " 1")
        setCurveMASHButton;

        }
        
        
        global proc curveButtonCMDS (string $nodeName, int $whichCMD) {
        if ($whichCMD == 1) {
        string $filter[] = `filterExpand -sm 9`;
        
        if (size($filter) > 0) {
        string $shape[] = `listRelatives -s $filter[0]`;    
        
        connectAttr -force ($shape[0]+".worldSpace[0]") ($nodeName+".inputCurve");
        print "Curve connected.";
        
        
        }
        else {
        warning "Please select a curve.";
        }
        }
        evalDeferred("updateAE " + $nodeName);
        }

        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
        self._curveLocation = [0,0,0,0]
        self._previousTime = om.MTime()
        self._initialised = False
        
    def compute(self, plug, dataBlock):

        if plug != mashCurve.outputArray:
            return om.kUnknownParameter
        
        #clear the out arrays - this node is untidy, and I'll probably re-organise it in the future
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        outRotResultArray = om.MVectorArray()
        outRotResultArray.clear()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashCurve.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashCurve.outputArray)
        
        #get the curve
        inCurve = dataBlock.inputValue( mashCurve.inputCurve )
        
        #get the interface values
        numberOfCalcs = dataBlock.inputValue(mashCurve.inIterations).asInt()
        timeStepValue = dataBlock.inputValue(mashCurve.timeStep).asFloat()
        timeSlideValue = dataBlock.inputValue(mashCurve.timeSlide).asFloat()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
        calcRotFlag = dataBlock.inputValue(mashCurve.calculateRotation).asBool()
        stopAtEndFlag = dataBlock.inputValue(mashCurve.stopAtEnd).asBool()
        EnvelopeValue = dataBlock.inputValue(mashCurve.aEnvelope).asFloat()
        EnvelopeRandValue = dataBlock.inputValue(mashCurve.aRandEnvelope).asFloat()
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        time = currentTime.value()
        
        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        inRotDArray = om.MVectorArray()
        inRotDArray.clear()
        inRotDArrData = dataBlock.inputValue(self.inRotationPP).data()
        inRotDArrFn = om.MFnVectorArrayData(inRotDArrData)
        inRotDArray = inRotDArrFn.array()
        
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outputArray
        outDData = dataBlock.outputValue(mashCurve.outputArray)
        nData = om.MFnVectorArrayData()
        
        #Set the output Rotation Array
        outRotDData = dataBlock.outputValue(mashCurve.outRotationPP)
        nRotData = om.MFnVectorArrayData()
        
        #Get the time
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.inputCurve)
        if plugMessage.isConnected():
            curveFn = om.MFnNurbsCurve(inCurve.asNurbsCurve())
            
            #not using api 2.0 so do have to something deranged to get doubles...
            min_util = om.MScriptUtil()
            min_util.createFromDouble(0)
            min_ptr = min_util.asDoublePtr()
            
            max_util = om.MScriptUtil()
            max_util.createFromDouble(0)
            max_ptr = max_util.asDoublePtr()
            
            #and use to get the min and max params of the curve
            curveFn.getKnotDomain( min_ptr, max_ptr )
            
            #and get the actual value out of the pointer
            curveLength = om.MScriptUtil().getDouble(max_ptr)
            
            curveStep = (timeStepValue*curveLength)*2
            
            #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
            if inDArray.length() > numberOfCalcs:
                numberOfCalcs = inDArray.length()
            
            if not self._initialised:
                #empty the lists
                del self._curveLocation[:]
                
                for h in range(numberOfCalcs):
                    stepValue = (curveStep/numberOfCalcs)*h
                    testRead = stepValue
                    #goal = om.MFloatVector(testRead.x,testRead.y,testRead.z)
                    #goalArray.append(goal)
                    self._curveLocation.append(testRead)
                    
                self._previousTime = currentTime
                self._initialised = True
            
            # did you click the timeline and mess up the calculations? Reset!
            timeStep = currentTime.value() - self._previousTime.value()
            if timeStep > 1.0 or timeStep < 0.0:
                self._initialised = False
                self._previousTime = currentTime
                return
                    

            
            for i in range(numberOfCalcs):
                if (inDArray.length() > 0):
                    testRead = om.MVector(0.0,0.0,0.0)
                    testRead = inDArray[i]
                
                if (inRotDArray.length() > 0):
                    rotRead = om.MVector(0.0,0.0,0.0)
                    rotRead = inRotDArray[i]
                
                # write a pointless line of code so that some stuff I copy and pasted works off the bat :)
                index = i
                
                point = om.MPoint()
                random.seed(i)
                timeSlideAddition = self._curveLocation[i] + timeSlideValue
                stepValue = timeSlideAddition
                
                if stepValue >= curveLength:
                    if (stopAtEndFlag !=1):
                        timeSlideAddition = 0
                        stepValue = 0.01 # a value of 0 will cause an error
                    else:
                        timeSlideAddition = curveLength
                        stepValue = curveLength-0.01 # a value of 0 or curveLength will cause an error
                
                if timeStep != 0:
                    self._curveLocation[i] = timeSlideAddition
                
                curveFn.getPointAtParam(stepValue, point, om.MSpace.kWorld)
                
                #calculate the rotation
                if (calcRotFlag!=0):
                    curveFn.tangent(stepValue, om.MSpace.kWorld)
                    
                    if stepValue == 0:
                        stepValue = 0.01
                    crvNorm = om.MVector(0.0,0.0,0.0)
                    crvNorm = curveFn.normal(stepValue, om.MSpace.kWorld);
                    
                    crvTan = om.MVector(0.0,0.0,0.0)
                    crvTan = curveFn.tangent(stepValue, om.MSpace.kWorld);

                    crvCross = om.MVector(0.0,0.0,0.0)
                    crvCross = (crvNorm^crvTan);
        
                    matrix = om.MMatrix()
                    om.MScriptUtil.setDoubleArray(matrix[0], 0, crvTan.x)
                    om.MScriptUtil.setDoubleArray(matrix[0], 1, crvTan.y)
                    om.MScriptUtil.setDoubleArray(matrix[0], 2, crvTan.z)
                    om.MScriptUtil.setDoubleArray(matrix[1], 0, crvCross.x)
                    om.MScriptUtil.setDoubleArray(matrix[1], 1, crvCross.y)
                    om.MScriptUtil.setDoubleArray(matrix[1], 2, crvCross.z)
                    om.MScriptUtil.setDoubleArray(matrix[2], 0, crvNorm.x)
                    om.MScriptUtil.setDoubleArray(matrix[2], 1, crvNorm.y)
                    om.MScriptUtil.setDoubleArray(matrix[2], 2, crvNorm.z)
                    
                    #put the matrix into a transform matrix
                    mTransformMtx = om.MTransformationMatrix(matrix)

                    #get the euler rotations
                    eulerRot = mTransformMtx.eulerRotation()

                    angles = [math.degrees(angle) for angle in (eulerRot.x, eulerRot.y, eulerRot.z)]

                else:
                    angles = [0,0,0]
                        
                if (EnvelopeRandValue < 1):
                    randStepValue = (EnvelopeRandValue) * (EnvelopeRandValue) * (3 - 2 * (EnvelopeRandValue))
                    
                    #randomise the smoothstep so that each object starts and ends at a different time.
                    for x in range(0,20):
                        whichSmoother = random.randint(0,20)
                        if (whichSmoother == 0):
                            #finishes earlier
                            randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                        if (whichSmoother == 1):
                            #starts later
                            randStepValue = randStepValue * randStepValue;
                    
                    A = 1;
                    B = 0;
                    
                    #finish off the smoothstep
                    newRandStepValue = (A * randStepValue) + (B * (1 - randStepValue))
                else:
                    newRandStepValue = EnvelopeRandValue

                stepResult = 1
                if (stepEnvelopeValue < 1):
                    #find the scale range for this i
                    step = (1.0/(numberOfCalcs+1))*(i+1)
                    minStep = (1.0/(numberOfCalcs+1))*i
                    
                    #are you step enviloping?
                    if stepEnvelopeValue <= step:
                        #normalise to get the correct scale value
                        stepResult = (stepEnvelopeValue-minStep)/(step-minStep)
                    else:
                        stepResult = 1.0
                    
                    if stepResult < step:
                        stepResult = 0.0
                
                inNormalArray = om.MFloatArray()
                inNormalArray.clear()
                
                #envilopes + and add the in array
                if (inDArray.length() > 0):
                    resultX = testRead.x+(point.x*EnvelopeValue)*newRandStepValue*stepResult
                    resultY = testRead.y+(point.y*EnvelopeValue)*newRandStepValue*stepResult
                    resultZ = testRead.z+(point.z*EnvelopeValue)*newRandStepValue*stepResult
                else:
                    resultX = (point.x*EnvelopeValue)*newRandStepValue
                    resultY = (point.y*EnvelopeValue)*newRandStepValue
                    resultZ = (point.z*EnvelopeValue)*newRandStepValue
                                
                if (inRotDArray.length() > 0):
                    angles[0] += rotRead.x
                    angles[1] += rotRead.y
                    angles[2] += rotRead.z
                
            
                thisResult = om.MVector(resultX,resultY,resultZ)
                outResultArray.append(thisResult)
                rotResult = om.MVector(angles[0],angles[1],angles[2])
                outRotResultArray.append(rotResult)
            
        else:
            for i in range(numberOfCalcs):
                angles = [0,0,0]
                if (inDArray.length() > 0):
                    testRead = om.MVector(0.0,0.0,0.0)
                    testRead = inDArray[i]
                    resultX = testRead.x
                    resultY = testRead.y
                    resultZ = testRead.z  
                else:
                    resultX = 0
                    resultY = 0
                    resultZ = 0
                    
                if (inRotDArray.length() > 0):
                    angles[0] += rotRead.x
                    angles[1] += rotRead.y
                    angles[2] += rotRead.z
                else:
                    angles[0] = 0
                    angles[1] = 0
                    angles[2] = 0
                
                rotResult = om.MVector(angles[0],angles[1],angles[2])
                thisResult = om.MVector(resultX,resultY,resultZ)
                outResultArray.append(thisResult)
                outRotResultArray.append(rotResult)
        
        self._previousTime = om.MTime(currentTime)
        nRotDataObj = nRotData.create(outRotResultArray)
        outRotDData.setMObject(nRotDataObj)
        
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashCurve() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashCurve.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashCurve.addAttribute( mashCurve.inputArray)
    
    tAttr = om.MFnTypedAttribute()
    mashCurve.inRotationPP = tAttr.create("inRotationPP", "inRot", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashCurve.addAttribute( mashCurve.inRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(0)
    mashCurve.addAttribute ( mashCurve.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashCurve.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(0)
    tAttr.setUsesArrayDataBuilder(True)
    mashCurve.addAttribute( mashCurve.outputArray )
    
    tAttr = om.MFnTypedAttribute()
    mashCurve.outRotationPP = tAttr.create("outRotationPP", "ourRotPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(0)
    tAttr.setUsesArrayDataBuilder(True)
    mashCurve.addAttribute( mashCurve.outRotationPP )
    
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.timeStep = nAttr.create ( "timeStep", "ts", om.MFnNumericData.kFloat, 0.1 )
    nAttr.setHidden(0)
    nAttr.setMax(1)
    nAttr.setMin(0)
    nAttr.setStorable(1)
    mashCurve.addAttribute ( mashCurve.timeStep )
    mashCurve.attributeAffects( mashCurve.timeStep, mashCurve.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.timeSlide = nAttr.create ( "timeSlide", "tsli", om.MFnNumericData.kFloat, 0.01 )
    nAttr.setHidden(0)
    nAttr.setMax(1)
    nAttr.setMin(0)
    nAttr.setStorable(1)
    mashCurve.addAttribute ( mashCurve.timeSlide )
    mashCurve.attributeAffects( mashCurve.timeSlide, mashCurve.outputArray)
    
    tAttr = om.MFnTypedAttribute()
    mashCurve.inputCurve = tAttr.create( "inputCurve", "ic", om.MFnNurbsCurveData.kNurbsCurve)
    mashCurve.addAttribute ( mashCurve.inputCurve )

    uAttr = om.MFnUnitAttribute()
    mashCurve.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashCurve.addAttribute(mashCurve.aTime)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.aEnvelope = nAttr.create("Envelope", "env", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashCurve.addAttribute( mashCurve.aEnvelope )
    mashCurve.attributeAffects( mashCurve.aEnvelope, mashCurve.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.aRandEnvelope = nAttr.create("randEnvelope", "raEn", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashCurve.addAttribute( mashCurve.aRandEnvelope )
    mashCurve.attributeAffects( mashCurve.aRandEnvelope, mashCurve.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashCurve.addAttribute( mashCurve.aStepEnvelope )
    mashCurve.attributeAffects( mashCurve.aStepEnvelope, mashCurve.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.calculateRotation = nAttr.create ( "calculateRotation", "calRot", om.MFnNumericData.kBoolean, 1 )
    mashCurve.addAttribute ( mashCurve.calculateRotation )
    mashCurve.attributeAffects( mashCurve.calculateRotation, mashCurve.outRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashCurve.stopAtEnd = nAttr.create ( "stopAtEnd", "sae", om.MFnNumericData.kBoolean, 0 )
    mashCurve.addAttribute ( mashCurve.stopAtEnd )
    mashCurve.attributeAffects( mashCurve.stopAtEnd, mashCurve.outputArray)
    
    mashCurve.attributeAffects( mashCurve.timeStep, mashCurve.outRotationPP)
    mashCurve.attributeAffects( mashCurve.timeSlide, mashCurve.outRotationPP)
    mashCurve.attributeAffects( mashCurve.aRandEnvelope, mashCurve.outRotationPP)
    mashCurve.attributeAffects( mashCurve.inputArray, mashCurve.outRotationPP)
    mashCurve.attributeAffects(mashCurve.aTime, mashCurve.outRotationPP)
    mashCurve.attributeAffects(mashCurve.aTime, mashCurve.outputArray)
    mashCurve.attributeAffects( mashCurve.inputArray, mashCurve.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashCurveId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashCurveId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    