import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Delay"

mashDelayId = om.MTypeId(0x0011BE07)

class mashDelay( omMPx.MPxNode ):
    
    inputs = om.MObject()
    outputArray = om.MObject()
    
    #AETemplate embedded in the .py file, get me.
    mel.eval('''
        global proc AEMASH_DelayTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Delay" "Delay";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "MASH Delay" -collapse 0;
        
        
        editorTemplate  -label "Time Step" -addControl "timeStep";
        editorTemplate  -label "Time Variance" -addControl "timeVariance";
        editorTemplate  -label "Time Offset" -addControl "offset";
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate -ccu "createInputs_MASH" "editInputs_MASH" "inputs" "Leader Object";
        
        editorTemplate -callCustom "AEaddLeaderButtons" "AEaddLeaderButtonsEdit" "";
        //editorTemplate  -label "Inputs" -addControl "inputs";
        editorTemplate  -label "Time" -addControl "time";
        
        editorTemplate -suppress "inIterations";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        global proc AEaddLeaderButtons ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Connect Leader"  -c ("delayButtonCMDS " + $nodeName[0] + " 1")
        connLeadMASHButton;
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        
        global proc AEaddLeaderButtonsEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("distButtonCMDS " + $nodeName[0] + " 1")
        connLeadMASHButton;
        
        }
        
        global proc delayButtonCMDS (string $nodeName, int $whichCMD)
        {
        if ($whichCMD == 1)
        {
        string $obj[] = `ls -sl -tr`;
        
        if (size($obj) > 0) {
        
        
        connectAttr -force ($obj[0]+".translate") ($nodeName+".inputs");
        
        print "Translate connected.";
        }
        else {
        warning "Please select a translate.";
        }
        }
        evalDeferred("updateAE " + $nodeName);
        }
        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):

        if plug != mashDelay.outputArray:
            return om.kUnknownParameter
        
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashDelay.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashDelay.outputArray)

        numberOfCalcs = dataBlock.inputValue(mashDelay.inIterations).asInt()
        timeStepValue = dataBlock.inputValue(mashDelay.timeStep).asFloat()
        timeVarianceValue = dataBlock.inputValue(mashDelay.timeVar).asFloat()
        timeOffsetValue = dataBlock.inputValue(mashDelay.timeOffset).asFloat()
        inputValues = dataBlock.inputValue(mashDelay.inputs).asFloat3()
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        time = currentTime.value()
        
        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outputArray
        outDData = dataBlock.outputValue(mashDelay.outputArray)
        nData = om.MFnVectorArrayData()
        
        #Get the time
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
                
        for i in range(numberOfCalcs):
            if (inDArray.length() > 0):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
            
            # Get the element index
            index = i
            
            random.seed(i)
            variation = random.uniform(0.0,timeVarianceValue)
            
            stepValue = (timeStepValue/numberOfCalcs)*i
            calcTime = (time-variation)-stepValue-timeOffsetValue
            
            #get the results at the previous time
            ctx = om.MDGContext(om.MTime(calcTime))
            thisNode = self.thisMObject()
            fnThisNode = om.MFnDependencyNode ( thisNode ) 
            plugX = fnThisNode.findPlug( 'inputs0' )
            plugY = fnThisNode.findPlug( 'inputs1' )
            plugZ = fnThisNode.findPlug( 'inputs2' )
            
            outX = plugX.asFloat(ctx)
            outY = plugY.asFloat(ctx)
            outZ = plugZ.asFloat(ctx)
            
            inNormalArray = om.MFloatArray()
            inNormalArray.clear()
            
            
            #if there is an input array, add those values            
            if (inDArray.length() > 0):
                resultX = testRead.x+outX
                resultY = testRead.y+outY
                resultZ = testRead.z+outZ
                #
            else:
                resultX = outX
                resultY = outY
                resultZ = outZ

            #add the results to the out array   
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)

        #set the output
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashDelay() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashDelay.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashDelay.addAttribute( mashDelay.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashDelay.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashDelay.addAttribute ( mashDelay.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashDelay.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashDelay.addAttribute( mashDelay.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDelay.timeStep = nAttr.create ( "timeStep", "ts", om.MFnNumericData.kFloat, 10 )
    nAttr.setHidden(0)
    nAttr.setSoftMax(10)
    nAttr.setMin(0)
    nAttr.setStorable(1)
    mashDelay.addAttribute ( mashDelay.timeStep )
    
    nAttr = om.MFnNumericAttribute()
    mashDelay.timeOffset = nAttr.create ( "offset", "off", om.MFnNumericData.kFloat, 0 )
    nAttr.setHidden(0)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(10)
    nAttr.setStorable(1)
    mashDelay.addAttribute ( mashDelay.timeOffset )
    
    nAttr = om.MFnNumericAttribute()
    mashDelay.timeVar = nAttr.create ( "timeVariance", "tva", om.MFnNumericData.kFloat, 20 )
    nAttr.setHidden(0)
    nAttr.setSoftMax(50)
    nAttr.setMin(0)
    nAttr.setStorable(1)
    mashDelay.addAttribute ( mashDelay.timeVar ) 
    
    nAttr = om.MFnNumericAttribute()
    mashDelay.inputs = nAttr.create ( "inputs", "in", om.MFnNumericData.k3Float, 0 )
    nAttr.setHidden(0)
    nAttr.setStorable(1)
    mashDelay.addAttribute ( mashDelay.inputs )
    
    uAttr = om.MFnUnitAttribute()
    mashDelay.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashDelay.addAttribute(mashDelay.aTime)
    mashDelay.attributeAffects(mashDelay.timeOffset, mashDelay.outputArray)
    mashDelay.attributeAffects(mashDelay.aTime, mashDelay.outputArray)
    mashDelay.attributeAffects( mashDelay.timeVar, mashDelay.outputArray)
    mashDelay.attributeAffects( mashDelay.timeStep, mashDelay.outputArray)
    mashDelay.attributeAffects( mashDelay.inputs, mashDelay.outputArray)
    mashDelay.attributeAffects( mashDelay.inputArray, mashDelay.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashDelayId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashDelayId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    