# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Distribute"
mashDistributeId = om.MTypeId(0x0011BEC3)

class mashDistribute( omMPx.MPxNode ):
    
    inputs = om.MObject()
    outputArray = om.MObject()
    inputArray = om.MObject()
    inputMesh = om.MObject()
    inIterations = om.MObject()
    
    #AETemplate in the .py, where it should be
    mel.eval('''
        global proc AEMASH_DistributeTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Distribute" "Distribute";

        editorTemplate -beginScrollLayout;

        editorTemplate -callCustom "AEarrangeMenu" "AEarrangeMenuEdit" "arrangement" $nodeName "amplitudeX" "amplitudeY" "amplitudeZ";

        editorTemplate -addSeparator;

        editorTemplate  -label "Strength" -addControl "envelope";
        editorTemplate  -label "Random Strength" -addControl "randomDistribute";
        editorTemplate  -label "Step Strength" -addControl "StepEnvelope";
        editorTemplate -addSeparator;
        editorTemplate  -label "Distribute Gap" -addControl "DistributeGap";
        editorTemplate  -label "Random Seed" -addControl "seed";
        editorTemplate -suppress "inIterations";
        editorTemplate -addSeparator;
        editorTemplate  -label "Offsets" -addControl "inputs";
        
        AEaddRampControl ($nodeName+".biasRamp");
        
        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Echo Settings (Instancer Mode Only)" -collapse 1;
        editorTemplate  -label "Echos" -addControl "numEchos";
        editorTemplate -addSeparator;
        editorTemplate  -label "Echo Offset X" -addControl "echoOffsetX";
        editorTemplate  -label "Echo Offset Y" -addControl "echoOffsetY";
        editorTemplate  -label "Echo Offset Z" -addControl "echoOffsetZ";
        editorTemplate -addSeparator;
        editorTemplate  -label "Echo Slide" -addControl "echoSlide";
        editorTemplate  -label "Echo Time Slide" -addControl "echoTimeSlide";
        editorTemplate -addSeparator;
        editorTemplate  -label "Echo Rotation" -addControl "echoRotation";
        editorTemplate  -label "Echo Rotation Variation" -addControl "echoRotationVar";
        editorTemplate  -label "Echo Scale" -addControl "echoScale";
        editorTemplate  -label "Echo Scale Variation" -addControl "echoScaleVar";
        
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Sphere Distribution Options" -collapse 1;

        editorTemplate  -label "Animation Speed" -addControl "noiseFrequency";     
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        
        editorTemplate  -label "Time" -addControl "time";       
        editorTemplate -ccu "createInputs_MASH" "editInputs_MASH" "inputMesh" "Input Mesh";

        editorTemplate -callCustom "AEaddMeshButtons" "AEaddMeshButtonsEdit" "";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        
        global proc AEarrangeMenu( string $attrName, string $nodeName, string $xName, string $yName, string $zName)
        {
        string $name[];
        tokenize($attrName, ".", $name);
        
        rowColumnLayout -nc 1 -cw 1 390 -cal 1 "center" distLayout_MASH;
        
        string $cmd = ("changeDistControls " + $attrName + " " + $nodeName);
        optionMenu -label "Arrangement Type:" -w 390 -cc $cmd arrangeTypeMenu;
        menuItem -label "Linear";
        menuItem -label "Radial";
        menuItem -label "Spherical";
        menuItem -label "Mesh Vertex";
        menuItem -label "inPositionPP";
        
        setParent..;
        
        int $arrValue = `getAttr $attrName`;
        
        if ($arrValue == 1) {
        optionMenu -e -v "Linear" arrangeTypeMenu;
        }
        if ($arrValue == 2) {
        optionMenu -e -v "Radial" arrangeTypeMenu;
        }
        if ($arrValue == 3) {
        optionMenu -e -v "Spherical" arrangeTypeMenu;
        }
        if ($arrValue == 4) {
        optionMenu -e -v "Mesh Vertex" arrangeTypeMenu;
        }
        if ($arrValue == 5) {
        optionMenu -e -v "inPositionPP" arrangeTypeMenu;
        }
        
        separator -w 390 -h 10 -hr 1 -st "in" distSep;
        
        floatSliderGrp -label "Amplitude X"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampXslider;
        
        connectControl ampXslider $xName;
        
        floatSliderGrp -label "Amplitude Y"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampYslider;
        
        connectControl ampYslider $yName;
        
        floatSliderGrp -label "Amplitude Z"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampZslider;
        
        connectControl ampZslider $zName;
        
        string $menu = `optionMenu -q -v arrangeTypeMenu`;
        
        if ($menu == "Linear") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Amplitude X" ampXslider;
        floatSliderGrp -e -l "Amplitude Y" ampYslider;
        floatSliderGrp -e -l "Amplitude Z" ampZslider;
        setAttr ($name[0]+".arrangement") 1;
        
        }
        else if ($menu == "Radial") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Radius" ampXslider;
        floatSliderGrp -e -l "Angle" ampYslider;
        floatSliderGrp -e -l "Z Step" ampZslider;
        setAttr ($name[0]+".arrangement") 2;
        
        }
        else if ($menu == "Spherical") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Angle X" ampXslider;
        floatSliderGrp -e -l "Angle Y" ampYslider;
        floatSliderGrp -e -l "Radius" ampZslider;
        setAttr ($name[0]+".arrangement") 3;
        
        }
        else if ($menu == "Mesh Vertex"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 4;
        
        }
        else if ($menu == "inPositionPP"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 5;
        
        }
        
        }
        
        
        global proc AEarrangeMenuEdit( string $attrName, string $nodeName, string $xName, string $yName, string $zName )
        {
        string $name[];
        tokenize($attrName, ".", $name);
        
        deleteUI arrangeTypeMenu; 
        deleteUI ampXslider;
        deleteUI ampYslider;
        deleteUI ampZslider;
        deleteUI distSep;
        deleteUI distLayout_MASH;

        rowColumnLayout -nc 1 -cw 1 390 -cal 1 "center" distLayout_MASH;
        
        string $cmd = ("changeDistControls " + $attrName + " " + $nodeName);
        optionMenu -label "Arrangement Type:" -w 390 -cc $cmd arrangeTypeMenu;
        menuItem -label "Linear";
        menuItem -label "Radial";
        menuItem -label "Spherical";
        menuItem -label "Mesh Vertex";
        menuItem -label "inPositionPP";
        
        setParent..;
        
        separator -w 390 -h 10 -hr 1 -st "in" distSep;
        
        floatSliderGrp -label "Amplitude X"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampXslider;
        
        connectControl ampXslider $xName;
        
        floatSliderGrp -label "Amplitude Y"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampYslider;
        
        connectControl ampYslider $yName;
        
        floatSliderGrp -label "Amplitude Z"
        -field true
        -minValue -90.0
        -maxValue 90.0
        -fieldMinValue -100000.0
        -fieldMaxValue 100000.0
        -value 0
        ampZslider;
        
        connectControl ampZslider $zName;
        
        
        string $cmd = ("changeDistControls " + $attrName + " " + $nodeName);
        optionMenu -e -cc $cmd arrangeTypeMenu;

        int $arrValue = `getAttr $attrName`;
        
        if ($arrValue == 1) {
        optionMenu -e -v "Linear" arrangeTypeMenu;
        }
        if ($arrValue == 2) {
        optionMenu -e -v "Radial" arrangeTypeMenu;
        }
        if ($arrValue == 3) {
        optionMenu -e -v "Spherical" arrangeTypeMenu;
        }
        if ($arrValue == 4) {
        optionMenu -e -v "Mesh Vertex" arrangeTypeMenu;
        }
        if ($arrValue == 5) {
        optionMenu -e -v "inPositionPP" arrangeTypeMenu;
        }
        
        string $menu = `optionMenu -q -v arrangeTypeMenu`;
        
        if ($menu == "Linear") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Amplitude X" ampXslider;
        floatSliderGrp -e -l "Amplitude Y" ampYslider;
        floatSliderGrp -e -l "Amplitude Z" ampZslider;
        setAttr ($name[0]+".arrangement") 1;
        
        }
        else if ($menu == "Radial") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Radius" ampXslider;
        floatSliderGrp -e -l "Angle" ampYslider;
        floatSliderGrp -e -l "Z Step" ampZslider;
        setAttr ($name[0]+".arrangement") 2;
        
        }
        else if ($menu == "Spherical") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Angle X" ampXslider;
        floatSliderGrp -e -l "Angle Y" ampYslider;
        floatSliderGrp -e -l "Radius" ampZslider;
        setAttr ($name[0]+".arrangement") 3;
        
        }
        else if ($menu == "Mesh Vertex"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 4;
        
        }
        else if ($menu == "inPositionPP"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 5;
        
        }
     
        }
        
        global proc changeDistControls(string $attrName, string $nodeName)
        {
        string $name[];
        tokenize($attrName, ".", $name);
        
        string $menu = `optionMenu -q -v arrangeTypeMenu`;
  
        if ($menu == "Linear") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Amplitude X" ampXslider;
        floatSliderGrp -e -l "Amplitude Y" ampYslider;
        floatSliderGrp -e -l "Amplitude Z" ampZslider;
        setAttr ($name[0]+".arrangement") 1;
        
        }
        else if ($menu == "Radial") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Radius" ampXslider;
        floatSliderGrp -e -l "Angle" ampYslider;
        floatSliderGrp -e -l "Z Step" ampZslider;
        setAttr ($name[0]+".arrangement") 2;
        
        }
        else if ($menu == "Spherical") {
        floatSliderGrp -e -en 1 ampXslider;
        floatSliderGrp -e -en 1 ampYslider;
        floatSliderGrp -e -en 1 ampZslider;
        floatSliderGrp -e -l "Angle X" ampXslider;
        floatSliderGrp -e -l "Angle Y" ampYslider;
        floatSliderGrp -e -l "Radius" ampZslider;
        setAttr ($name[0]+".arrangement") 3;
        
        }
        else if ($menu == "Mesh Vertex"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 4;
        
        }
        else if ($menu == "inPositionPP"){
        
        floatSliderGrp -e -en 0 ampXslider;
        floatSliderGrp -e -en 0 ampYslider;
        floatSliderGrp -e -en 0 ampZslider;
        setAttr ($name[0]+".arrangement") 5;
        
        }
        
        }
        
        
        global proc AEaddMeshButtons ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Connect Mesh"  -c ("distButtonCMDS " + $nodeName[0] + " 1")
        distMeshMASHButton;
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        
        global proc AEaddMeshButtonsEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("distButtonCMDS " + $nodeName[0] + " 1")
        distMeshMASHButton;
        
        }
        
        global proc distButtonCMDS (string $nodeName, int $whichCMD)
        {
        if ($whichCMD == 1)
        {
        string $filter[] = `filterExpand -sm 12`;
        
        if (size($filter) > 0) {
        string $shape[] = `listRelatives -s $filter[0]`;
        
        connectAttr -force ($shape[0]+".worldMesh") ($nodeName+".inputMesh");

        print "Mesh connected.";
        }
        else {
        warning "Please select a Mesh.";
        }
        }
        evalDeferred("updateAE " + $nodeName);
        }
        
        ''')
    
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):

        if plug != mashDistribute.outputArray:
            return om.kUnknownParameter
        
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #get the ramp, ( pilfered wholesale from this clever Chap http://rosenio.com/blog/?p=95 )
        thisNode = self.thisMObject()
        rampPlug = om.MPlug( thisNode, mashDistribute.aBiasRamp )
        rampPlug.setAttribute(mashDistribute.aBiasRamp)
        RampPlug01 = om.MPlug(rampPlug.node(), rampPlug.attribute())
    
        #Get the atrributes as MRampAttribute
        myRamp = om.MRampAttribute(RampPlug01.node(), RampPlug01.attribute())
                            
        # Get the input handle
        dataHandle = dataBlock.inputValue(mashDistribute.inputs).asFloat3()
        
        #get the interface values
        valueX = dataBlock.inputValue(mashDistribute.aAmplitudeX).asFloat()
        valueY = dataBlock.inputValue(mashDistribute.aAmplitudeY).asFloat()
        valueZ = dataBlock.inputValue(mashDistribute.aAmplitudeZ).asFloat()
        envelopeValue = dataBlock.inputValue(mashDistribute.aEnvelope).asFloat()
        seedValue = dataBlock.inputValue(mashDistribute.aSeed).asInt()
        DistributeGapValue = dataBlock.inputValue(mashDistribute.aDistributeGap).asInt()
        arrangementStyle = dataBlock.inputValue(mashDistribute.aArrangement).asInt()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
        
        #echo values
        echoValue = dataBlock.inputValue(mashDistribute.numEchos).asInt()
        echoAmplitudeXval = dataBlock.inputValue(mashDistribute.aEchoAmplitudeX).asFloat()
        echoAmplitudeYval = dataBlock.inputValue(mashDistribute.aEchoAmplitudeY).asFloat()
        echoAmplitudeZval = dataBlock.inputValue(mashDistribute.aEchoAmplitudeZ).asFloat()
        echoOffsetXval = dataBlock.inputValue(mashDistribute.aEchoOffsetX).asFloat()
        echoOffsetYval = dataBlock.inputValue(mashDistribute.aEchoOffsetY).asFloat()
        echoOffsetZval = dataBlock.inputValue(mashDistribute.aEchoOffsetZ).asFloat()
        echoRotationVal = dataBlock.inputValue(mashDistribute.aEchoRotation).asFloat()
        echoScaleVal = dataBlock.inputValue(mashDistribute.aEchoScale).asFloat()
        echoSlideVal = dataBlock.inputValue(mashDistribute.aEchoSlide).asFloat()
        echoScaleVarVal = dataBlock.inputValue(mashDistribute.aEchoScaleVar).asFloat()
        echoRotVarVal = dataBlock.inputValue(mashDistribute.aEchoRotationVar).asFloat()
        echoTimeSlideVal = dataBlock.inputValue(mashDistribute.aEchoTimeSlide).asFloat()
        
        sphereFrequency = dataBlock.inputValue(mashDistribute.aNoiseFrequency).asFloat()
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashDistribute.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashDistribute.outputArray)
        
        #how many times to run the code
        numberOfCalcs = dataBlock.inputValue(mashDistribute.inIterations).asInt()

        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        

        
        
        inPosDArray = om.MVectorArray()
        inPosDArray.clear()
        inPosDArrData = dataBlock.inputValue(self.inPositionPP).data()
        inPosDArrFn = om.MFnVectorArrayData(inPosDArrData)
        inPosDArray = inPosDArrFn.array()
        
        testRead = om.MVector(0.0,0.0,0.0)
        
        #get the falloff object position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
            
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        if inputs.length() > 0:
            
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
        
        vts = om.MPointArray()
        
        #create a dummy list, and fill with inane data to keep Maya happy
        testList = ['A', 'B', 'C', 'D']
        #get the input mesh for vertex placement if we're in vertex mode
        if (arrangementStyle == 4):
            #get the mesh
            inputData = dataBlock.inputValue(mashDistribute.inputMesh)
            inMesh = inputData.data()
            meshFS = om.MFnMesh()
            meshFS = om.MFnMesh(inMesh)
            #number of vertices
            oriVertCount = meshFS.numVertices()
            #if there are any vertices, add them to an array
            if (oriVertCount >0):
                meshFS.getPoints(vts,om.MSpace.kWorld)
                vtsSize = vts.length()
                #clear the dummy list
                del testList[:]
                #add the vertex positions to the list, then shuffle the list order so objects get put on random vertices.
                for i in range(vtsSize):
                    testList.append(vts[i])
                random.shuffle(testList)
        
        #Setup the outputArray
        outDData = dataBlock.outputValue(mashDistribute.outputArray)
        nData = om.MFnVectorArrayData()
        
        #iterator for the mesh mode
        vertCycle = 0
        
        inCtxDArray = om.MVectorArray()
        inCtxDArray.clear()
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
        
        for e in range(echoValue+1):
            #if we're echoing
            if e > 0:
                #and we're shifting the echos in time
                if echoTimeSlideVal > 0:
                    #give the echo a random seed
                    random.seed(e)
                    #work out a random offset in time based on the slider and seed
                    randomOffset = random.uniform(0-echoTimeSlideVal, 0)
                    calcTime = currentTime.value() - randomOffset
                    #create the context
                    ctx = om.MDGContext(om.MTime(calcTime))
                    #find the plug
                    thisNode = self.thisMObject()
                    fnThisNode = om.MFnDependencyNode ( thisNode )
                    plug = fnThisNode.findPlug( 'inputArray' )
                    #MObject must be used here because .data doesn't work when accesing via plugs.
                    oData = plug.asMObject(ctx)#this loads the plug data at the ctx stated time.
                    inCtxDArrFn = om.MFnVectorArrayData(oData)
                    #is there any data in the array, if so, set the array, this step is ESSENTIAL as zero values are often returned (bug?) and this breaks everything.
                    if inCtxDArrFn.array().length() > 0:
                        inDArray = inCtxDArrFn.array()

            #run the calculations as many times as needed
            for i in range(numberOfCalcs):
                testRead = om.MVector(0.0,0.0,0.0)
                #if theres an in array, get the value
                if (inDArray.length() > 0):
                    testRead = inDArray[i]
                            
                #make extra work for myself by not just using i
                index = i
                
                inNormalArray = om.MFloatArray()
                inNormalArray.clear()
                
                #calculate the Distributegap index
                DistributeGapIndex = math.ceil(index/DistributeGapValue)
                
                #when there's more then one calculation, say there's one less
                if (numberOfCalcs > 1):
                    if (arrangementStyle==2):
                        numberOfOutputs = numberOfCalcs-1
                else:
                    numberOfOutputs = numberOfCalcs
                
                #where is this object onthe ramp
                rampPosition = (1.0/(numberOfCalcs-1))*DistributeGapIndex
                
                valAt_util = om.MScriptUtil()
                
                #Get the value from the ramp
                valAt_util.createFromDouble(1.0)
                valAtPtr = valAt_util.asFloatPtr()
                myRamp.getValueAtPosition(rampPosition, valAtPtr)
                
                #Get the value at pointer as float
                valAtPos = valAt_util.getFloat(valAtPtr)
                
                #get a fresh randDistribute value
                randDistributeValue = dataBlock.inputValue(mashDistribute.aRandomDistribute).asFloat()
            
                ePlus = e+10
                random.seed(index*seedValue*ePlus)
            
                #smoothDistribute the randDistributeValue
                if (randDistributeValue < 1):
                    randStepValue = (randDistributeValue) * (randDistributeValue) * (3 - 2 * (randDistributeValue))
                    
                    #randomise the smoothstep so that each object starts and ends at a different time.
                    for x in range(0,20):
                        whichSmoother = random.randint(0,20)
                        if (whichSmoother == 0):
                            #finishes earlier
                            randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                        if (whichSmoother == 1):
                            #starts later
                            randStepValue = randStepValue * randStepValue;
                    
                    A = 1;
                    B = 0;
                    
                    #finish off the smoothstep
                    newRandDistributeValue = (A * randStepValue) + (B * (1 - randStepValue))
                    
                else:
                    newRandDistributeValue = randDistributeValue
                
                if (arrangementStyle==2):
                    #radial
                    twoPi = 6.28318*(valueY/360)
                    angle = ((twoPi / (numberOfOutputs+1))*index) + (math.radians(echoRotationVal)*e)
                    DistributeZ = (0+((((valueZ*i)+(echoAmplitudeZval*e))*valAtPos)*newRandDistributeValue)*envelopeValue)
                    DistributeY = (((math.cos( angle  +(echoAmplitudeXval*e)) * (valueX+(echoAmplitudeYval*e)))*valAtPos)*newRandDistributeValue)*envelopeValue
                    DistributeX = (((math.sin( angle  +(echoAmplitudeXval*e)) * (valueX+(echoAmplitudeYval*e)))*valAtPos)*newRandDistributeValue)*envelopeValue
                   
                elif (arrangementStyle==3):
                    #spherical
                    t = currentTime.value()*(sphereFrequency*0.01)

                    twoPi = 6.28318
                    theta = (random.uniform(0.0, twoPi*(valueX/360)))+t
                    phi = (random.uniform(0.0, twoPi*(valueY/360)))+t
                    DistributeX = (((math.cos(theta) * math.sin(phi) * valueZ)*valAtPos)*newRandDistributeValue)*envelopeValue
                    DistributeY = (((math.sin(theta) * math.sin(phi) * valueZ)*valAtPos)*newRandDistributeValue)*envelopeValue
                    DistributeZ = (((math.cos(phi) * valueZ)*valAtPos)*randDistributeValue)*envelopeValue              

                elif (arrangementStyle==4):
                    #mesh
                    if (numberOfCalcs == vertCycle):
                        vertCycle = 0
                    DistributeX = ((testList[vertCycle].x)*newRandDistributeValue)*envelopeValue
                    DistributeY = ((testList[vertCycle].y)*newRandDistributeValue)*envelopeValue
                    DistributeZ = ((testList[vertCycle].z)*newRandDistributeValue)*envelopeValue
                    vertCycle += 1
                elif (arrangementStyle==5):
                    #inPositionPP
                    if (inPosDArray.length() >0):
                        posRead = om.MVector(0.0,0.0,0.0)
                        posRead = inPosDArray[i]
                        DistributeX = (posRead.x*newRandDistributeValue)*envelopeValue
                        DistributeY = (posRead.y*newRandDistributeValue)*envelopeValue
                        DistributeZ = (posRead.z*newRandDistributeValue)*envelopeValue
                    else:
                        om.MGlobal.displayWarning("Please connect an attribute to the inPositionPP Attribute")
                else:
                    #normal
                    DistributeX = ((valueX*valAtPos)*newRandDistributeValue)*envelopeValue#+(echoOffsetXval*e)
                    DistributeY = ((valueY*valAtPos)*newRandDistributeValue)*envelopeValue#+(echoOffsetYval*e)
                    DistributeZ = ((valueZ*valAtPos)*newRandDistributeValue)*envelopeValue#+(echoOffsetZval*e)
            
            
                stepResult = 1
                if (stepEnvelopeValue < 1):
                    #find the scale range for this i
                    step = (1.0/(numberOfCalcs+1))*(i+1)
                    minStep = (1.0/(numberOfCalcs+1))*i
                    
                    #are you step enviloping?
                    if stepEnvelopeValue <= step:
                        #normalise to get the correct scale value
                        stepResult = (stepEnvelopeValue-minStep)/(step-minStep)
                    else:
                        stepResult = 1.0
                    
                    if stepResult < step:
                        stepResult = 0.0
            
                #if i  < 2:
                #    print "max", step, "min", min, "slider", stepEnvelopeValue
            
                DistributeX = DistributeX*stepResult + dataHandle[0]
                DistributeY = DistributeY*stepResult + dataHandle[1]
                DistributeZ = DistributeZ*stepResult + dataHandle[2]
                
                #randomSeed should now only be different for each echo.
                random.seed(e+seedValue)
                #Rotate the echos with some fancy trig (well, it's fancy to me)
                if e > 0:
                    twoPi = 6.28318
                    step = (echoRotationVal/numberOfCalcs)
                    randomRot = random.uniform(0-echoRotVarVal, echoRotVarVal)
                    angle = (step+(randomRot*0.0174532925)) * e
                    s = math.sin(angle)
                    c = math.cos(angle)
                
                    if (inDArray.length() > 0):
                        #if there's an in array, add those values
                        inverseX = 0- (testRead.x + DistributeX)
                        inverseY = 0- (testRead.y + DistributeY)
                    else:
                        inverseX = 0 - DistributeX
                        inverseY = 0 - DistributeY
                    
                    xNew = inverseX * c - inverseY * s
                    yNew = inverseX * s + inverseY * c
                    
                    DistributeX = 0-xNew-testRead.x
                    DistributeY = 0-yNew-testRead.y

                
                #slide calculation
                totalSlideX = echoOffsetXval*echoValue
                totalSlideY = echoOffsetYval*echoValue
                totalSlideZ = echoOffsetZval*echoValue
                randomNumberX = random.uniform(0-totalSlideX,totalSlideX)
                randomNumberY = random.uniform(0-totalSlideY,totalSlideY)
                randomNumberZ = random.uniform(0-totalSlideZ,totalSlideZ)
                randomNumberX *= echoSlideVal
                randomNumberY *= echoSlideVal
                randomNumberZ *= echoSlideVal
                    
                #scale the echo values
                randomScale = random.uniform(0-echoScaleVarVal, echoScaleVarVal)
                scaleVariation = echoScaleVal+randomScale
                scaleStep = 1+(scaleVariation/numberOfCalcs)*(e+1)
                DistributeX = (DistributeX*scaleStep)+randomNumberX
                DistributeY = (DistributeY*scaleStep)+randomNumberY
                DistributeZ = (DistributeZ*scaleStep)+randomNumberZ

                resultX = DistributeX
                resultY = DistributeY
                resultZ = DistributeZ
            
                #using a falloff object?
                if useFalloffFlag==1:
                    if inputs.length() > 0:
                        if radius == 0:
                            radius = 0.001 #avoid divide by zero error
                        
                        #how far is our object from the sphere centre?
                        #fast approximate method, not really good enough.
                        '''
                        distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                        distanceSqrd *= 0.1
                        '''
                        #take echos into account
                        tx = testRead.x + (echoOffsetXval*e)
                        ty = testRead.y + (echoOffsetYval*e)
                        tz = testRead.z + (echoOffsetZval*e)
                        
                        #slow more accurate method
                        xd = tx - falloffObjPos.x
                        yd = ty - falloffObjPos.y
                        zd = tz - falloffObjPos.z
                        distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                        
                        
                        #linear falloff
                        falloff = min(1, max(0, distanceSqrd/radius))
                        
                        #normalise the result to take into account the minimum radius
                        if innerRadius > 0: #avoid divide by zero error
                            falloff = (falloff - innerRadius) / (1-innerRadius)
                        
                        #correct for times when we get a negative result (with an innerZone close to 1)
                        if falloff < 0:
                            falloff = 0
                        
                        #smooth step the linear falloff to make smooth falloff
                        A = 1;
                        B = 0;
                        
                        #trying to decide which of these smooth steps I prefer...
                        #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                        randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                        falloff = (A * randStepValue) + (B * (1 - randStepValue))
                        
                        if invertFalloffFlag !=1:
                            falloffEnvelope = 1-falloff
                        else:
                            falloffEnvelope = falloff
                        
                        resultX *= falloffEnvelope
                        resultY *= falloffEnvelope
                        resultZ *= falloffEnvelope
                            
        
                #add offsets after everything else so that they don't get enviloped.
                resultX += (echoOffsetXval*e)
                resultY += (echoOffsetYval*e)
                resultZ += (echoOffsetZval*e)
        
                if (inDArray.length() > 0):

                    resultX = testRead.x + resultX
                    resultY = testRead.y + resultY
                    resultZ = testRead.z + resultZ


                #set this result
                thisResult = om.MVector(resultX,resultY,resultZ)
                outResultArray.append(thisResult)        
    
        #set the values
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        
        #polish the plug for next time
        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashDistribute() )

def nodeInitializer():
    
    mashDistribute.aBiasRamp= om.MRampAttribute.createCurveRamp("biasRamp", "bRmp")
    
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.buffer = nAttr.create ( "buffer", "buf", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    nAttr.setWritable(0)
    nAttr.setReadable(0)
    nAttr.setStorable(0)
    mashDistribute.addAttribute ( mashDistribute.buffer )

    tAttr = om.MFnTypedAttribute()
    mashDistribute.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashDistribute.addAttribute( mashDistribute.outputArray )
    
    tAttr = om.MFnTypedAttribute()
    mashDistribute.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashDistribute.addAttribute( mashDistribute.inputArray)
    
    tAttr = om.MFnTypedAttribute()
    mashDistribute.inPositionPP = tAttr.create("inPositionPP", "inPPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashDistribute.addAttribute( mashDistribute.inPositionPP)
    mashDistribute.attributeAffects( mashDistribute.inPositionPP , mashDistribute.outputArray )
    
    nAttrB = om.MFnNumericAttribute()
    mashDistribute.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    mashDistribute.addAttribute ( mashDistribute.inIterations )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.inputs = nAttr.create ( "inputs", "in", om.MFnNumericData.k3Float, 0 )
    nAttr.setHidden(0)
    nAttr.setStorable(1)
    mashDistribute.addAttribute ( mashDistribute.inputs )

    nAttr = om.MFnNumericAttribute()
    mashDistribute.aSeed = nAttr.create("seed", "see", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashDistribute.addAttribute( mashDistribute.aSeed )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aAmplitudeX = nAttr.create("amplitudeX", "ampX", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-90)
    nAttr.setSoftMax(90)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aAmplitudeY = nAttr.create("amplitudeY", "ampY", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-90)
    nAttr.setSoftMax(90)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aAmplitudeZ = nAttr.create("amplitudeZ", "ampZ", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-90)
    nAttr.setSoftMax(90)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aDistributeGap = nAttr.create("DistributeGap", "stGp", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEnvelope = nAttr.create("envelope", "env", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEnvelope )
    mashDistribute.attributeAffects( mashDistribute.aEnvelope , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aStepEnvelope )
    mashDistribute.attributeAffects( mashDistribute.aStepEnvelope, mashDistribute.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aRandomDistribute = nAttr.create("randomDistribute", "rdSp", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aRandomDistribute )
    mashDistribute.attributeAffects( mashDistribute.aRandomDistribute , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoAmplitudeX = nAttr.create("echoAmplitudeX", "eampX", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-1)
    nAttr.setSoftMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoAmplitudeX )
    mashDistribute.attributeAffects( mashDistribute.aEchoAmplitudeX , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoAmplitudeY = nAttr.create("echoAmplitudeY", "eampY", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-1)
    nAttr.setSoftMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoAmplitudeY )
    mashDistribute.attributeAffects( mashDistribute.aEchoAmplitudeY , mashDistribute.outputArray )

    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoAmplitudeZ = nAttr.create("echoAmplitudeZ", "eampZ", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-1)
    nAttr.setSoftMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoAmplitudeZ )
    mashDistribute.attributeAffects( mashDistribute.aEchoAmplitudeZ , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoOffsetX = nAttr.create("echoOffsetX", "eoffX", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-10)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoOffsetX )
    mashDistribute.attributeAffects( mashDistribute.aEchoOffsetX , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoOffsetY = nAttr.create("echoOffsetY", "eoffY", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-10)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoOffsetY )
    mashDistribute.attributeAffects( mashDistribute.aEchoOffsetY , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoOffsetZ = nAttr.create("echoOffsetZ", "eoffZ", om.MFnNumericData.kFloat, -2.0)
    nAttr.setSoftMin(-10)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoOffsetZ )
    mashDistribute.attributeAffects( mashDistribute.aEchoOffsetZ , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoRotation = nAttr.create("echoRotation", "ecRot", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(90)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoRotation )
    mashDistribute.attributeAffects( mashDistribute.aEchoRotation , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoRotationVar = nAttr.create("echoRotationVar", "ecRotV", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(90)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoRotationVar )
    mashDistribute.attributeAffects( mashDistribute.aEchoRotationVar , mashDistribute.outputArray )

    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoScale= nAttr.create("echoScale", "ecSc", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-10)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoScale )
    mashDistribute.attributeAffects( mashDistribute.aEchoScale , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoScaleVar= nAttr.create("echoScaleVar", "ecScV", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-1)
    nAttr.setSoftMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoScaleVar )
    mashDistribute.attributeAffects( mashDistribute.aEchoScaleVar , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoSlide= nAttr.create("echoSlide", "ecSl", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(-1)
    nAttr.setSoftMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoSlide )
    mashDistribute.attributeAffects( mashDistribute.aEchoSlide , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aEchoTimeSlide= nAttr.create("echoTimeSlide", "ecTs", om.MFnNumericData.kFloat, 0.0)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(100)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aEchoTimeSlide )
    mashDistribute.attributeAffects( mashDistribute.aEchoTimeSlide , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.aNoiseFrequency = nAttr.create("noiseFrequency", "noFre", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setSoftMax(3)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.aNoiseFrequency )
    mashDistribute.attributeAffects( mashDistribute.aNoiseFrequency , mashDistribute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashDistribute.numEchos = nAttr.create("numEchos", "echo", om.MFnNumericData.kInt, 0)
    nAttr.setMin(0)
    nAttr.setSoftMax(10)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashDistribute.addAttribute( mashDistribute.numEchos )
    mashDistribute.attributeAffects( mashDistribute.numEchos , mashDistribute.outputArray )

    inTypedAttr = om.MFnTypedAttribute()
    mashDistribute.inputMesh = inTypedAttr.create("inputMesh", "inM", om.MFnData.kMesh)
    mashDistribute.addAttribute(mashDistribute.inputMesh)
    
    mashDistribute.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashDistribute.addAttribute(mashDistribute.falloffObject)
    mashDistribute.attributeAffects(mashDistribute.falloffObject, mashDistribute.outputArray)
    
    mAttr = om.MFnMessageAttribute()
    mashDistribute.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashDistribute.addAttribute( mashDistribute.falloffMessage )

    typedAttr = om.MFnTypedAttribute()
    enumAttr = om.MFnEnumAttribute()
    mashDistribute.aArrangement = enumAttr.create("arrangement", "rt", om.MFnData.kNumeric)
    enumAttr.addField('Normal', 1)
    enumAttr.addField('Radial', 2)
    enumAttr.addField('Spherical', 3)
    enumAttr.addField('Mesh Vertex', 4)
    enumAttr.addField('inPositionPP', 5)
    mashDistribute.randomeData = typedAttr.create("arrangementTypeData", "atd", om.MFnData.kString)
    mashDistribute.addAttribute ( mashDistribute.aArrangement )
    mashDistribute.attributeAffects( mashDistribute.aArrangement, mashDistribute.outputArray)
    
    uAttr = om.MFnUnitAttribute()
    mashDistribute.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashDistribute.addAttribute(mashDistribute.aTime)
    mashDistribute.attributeAffects(mashDistribute.aTime, mashDistribute.outputArray)
    
    mashDistribute.addAttribute( mashDistribute.aBiasRamp )
    mashDistribute.addAttribute( mashDistribute.aAmplitudeX )
    mashDistribute.addAttribute( mashDistribute.aAmplitudeY )
    mashDistribute.addAttribute( mashDistribute.aAmplitudeZ )
    
    mashDistribute.addAttribute( mashDistribute.aDistributeGap )
    
    mashDistribute.attributeAffects( mashDistribute.aSeed , mashDistribute.outputArray )
    mashDistribute.attributeAffects( mashDistribute.aBiasRamp , mashDistribute.outputArray )

    mashDistribute.attributeAffects( mashDistribute.inputArray, mashDistribute.outputArray)
    mashDistribute.attributeAffects( mashDistribute.inputs, mashDistribute.outputArray)
    mashDistribute.attributeAffects( mashDistribute.aDistributeGap, mashDistribute.outputArray)
    
    mashDistribute.attributeAffects( mashDistribute.aAmplitudeX, mashDistribute.outputArray)
    mashDistribute.attributeAffects( mashDistribute.aAmplitudeY, mashDistribute.outputArray)
    mashDistribute.attributeAffects( mashDistribute.aAmplitudeZ, mashDistribute.outputArray)
    mashDistribute.attributeAffects(mashDistribute.inputMesh, mashDistribute.outputArray)


def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashDistributeId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashDistributeId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
