# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/
# with some serious help from this tutorial http://www.fevrierdorian.com/blog/post/2010/02/12/Creating-custom-locator-with-Maya-s-Python-API

import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.OpenMayaRender as omRender
import maya.OpenMayaUI as omUI
import maya.cmds as cmds
import math
import sys
import maya.mel as mel

nodeType = "MASH_Falloff"
nodeId = om.MTypeId(0x0011BE51)
glRenderer = omRender.MHardwareRenderer.theRenderer()
glFT = glRenderer.glFunctionTable()

class mashFalloff(omMPx.MPxLocatorNode):	

    aColor = om.MObject()
    aRadius = om.MObject()

    mel.eval('''
        global proc AEMASH_FalloffTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Falloff" "Falloff";

        editorTemplate -beginScrollLayout;
        editorTemplate -beginLayout "Falloff Node" -collapse 0;
        
        editorTemplate  -label "Enable" -addControl "useFalloff";
        editorTemplate  -label "Invert Falloff" -addControl "invertFalloff";
        editorTemplate -addSeparator;
        editorTemplate  -label "Radius" -addControl "radius";
        editorTemplate  -label "Inner Zone" -addControl "innerRadius";
        editorTemplate -addSeparator;
        editorTemplate  -label "Colour" -addControl "shapeColor";

        editorTemplate -endLayout;
        
        AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
                
        ''')

	# init
    def __init__(self):		
        omMPx.MPxLocatorNode.__init__(self)
	
    
    # post constructor required to set the name of the locator properly
    def postConstructor(self):
        this_object	= self.thisMObject()
        fnThisNode = om.MFnDependencyNode(this_object)
        fnThisNode.setName("MASH_FalloffShape#")
    
	# nothing to do here
    def compute(self, plug, datahandle):						
        return om.kUnknownParameter
		
	# draw
    def draw(self, view, path, style, status):
		
		#get the interface values 
        this_object		= self.thisMObject()
        fnThisNode	= om.MFnDependencyNode(this_object)
        locName		= fnThisNode.name()

        plug = om.MPlug(this_object, self.aRadius)
        radius = plug.asFloat()
        
        plug = om.MPlug(this_object, self.aInnerRadius)
        innerRadius = plug.asFloat()

        plug = om.MPlug(this_object, self.aColor)
        col = plug.asMObject()
        numDataFn = om.MFnNumericData(col)

        #rgb values as pointers
        rParam = om.MScriptUtil()
        rParam.createFromDouble(0.0)
        rPtr = rParam.asFloatPtr()

        gParam = om.MScriptUtil()
        gParam.createFromDouble(0.0)
        gPtr = gParam.asFloatPtr()

        bParam = om.MScriptUtil()
        bParam.createFromDouble(0.0)
        bPtr = bParam.asFloatPtr()

        numDataFn.getData3Float(rPtr, gPtr, bPtr)
        
        #colour as floats
        colour = om.MFloatVector(om.MScriptUtil(rPtr).asFloat(), om.MScriptUtil(gPtr).asFloat(), om.MScriptUtil(bPtr).asFloat())
		
        # start openGL
        view.beginGL()
        if style == omUI.M3dView.kFlatShaded or style == omUI.M3dView.kGouraudShaded:
            glFT.glPushAttrib(omRender.MGL_CURRENT_BIT)
            glFT.glPopAttrib()
		
        # set colour
        if status == omUI.M3dView.kDormant:
            glFT.glColor3f(colour[0],colour[1],colour[2])

        # line width
        glFT.glLineWidth(1)
        glFT.glBegin(omRender.MGL_LINES)

        #draw outer radius lines
        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(radius*math.cos(0), radius*math.sin(0), 0.0)
                glFT.glVertex3f(radius*math.cos(0), radius*math.sin(0), 0.0)
            else:
                glFT.glTexCoord3f(radius*math.cos(rad), radius*math.sin(rad), 0.0)
                glFT.glVertex3f(radius*math.cos(rad), radius*math.sin(rad), 0.0)
                glFT.glVertex3f(radius*math.cos(radPlus), radius*math.sin(radPlus), 0.0)
        
        
        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(0.0, radius*math.sin(0), radius*math.cos(0))
                glFT.glVertex3f(0.0, radius*math.sin(0),radius*math.cos(0))
            
            else:
                glFT.glTexCoord3f(0.0, radius*math.sin(rad), radius*math.cos(rad))
                glFT.glVertex3f(0.0, radius*math.sin(rad), radius*math.cos(rad))
                glFT.glVertex3f(0.0, radius*math.sin(radPlus), radius*math.cos(radPlus))
    
        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(radius*math.sin(0), 0.0, radius*math.cos(0))
                glFT.glVertex3f(radius*math.sin(0), 0.0,radius*math.cos(0))
            else:
                glFT.glTexCoord3f(radius*math.sin(rad), 0.0, radius*math.cos(rad))
                glFT.glVertex3f(radius*math.sin(rad), 0.0, radius*math.cos(rad))
                glFT.glVertex3f(radius*math.sin(radPlus), 0.0, radius*math.cos(radPlus))

        
        innerRadiusValue = radius*innerRadius
        
        #draw inner radius with dots
        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(innerRadiusValue*math.cos(0), innerRadiusValue*math.sin(0), 0.0)
                glFT.glVertex3f(innerRadiusValue*math.cos(0), innerRadiusValue*math.sin(0), 0.0)
            else:
                glFT.glTexCoord3f(innerRadiusValue*math.cos(rad), innerRadiusValue*math.sin(rad), 0.0)
                glFT.glVertex3f(innerRadiusValue*math.cos(rad), innerRadiusValue*math.sin(rad), 0.0)
                #glFT.glVertex3f(innerRadiusValue*math.cos(radPlus), innerRadiusValue*math.sin(radPlus), 0.0) #solid lines with this line uncommented

        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(0.0, innerRadiusValue*math.sin(0), innerRadiusValue*math.cos(0))
                glFT.glVertex3f(0.0, innerRadiusValue*math.sin(0),innerRadiusValue*math.cos(0))
            
            else:
                glFT.glTexCoord3f(0.0, innerRadiusValue*math.sin(rad), innerRadiusValue*math.cos(rad))
                glFT.glVertex3f(0.0, innerRadiusValue*math.sin(rad), innerRadiusValue*math.cos(rad))
                #glFT.glVertex3f(0.0, innerRadiusValue*math.sin(radPlus), innerRadiusValue*math.cos(radPlus))

        for i in range(0,360):
            rad = (i*2*math.pi)/360;
            radPlus = ((i+1)*2*math.pi)/360;
            glFT.glNormal3f(0.0, 0.0, 1.0)
            if (i == 360):
                glFT.glTexCoord3f(innerRadiusValue*math.sin(0), 0.0, innerRadiusValue*math.cos(0))
                glFT.glVertex3f(innerRadiusValue*math.sin(0), 0.0,innerRadiusValue*math.cos(0))
            else:
                glFT.glTexCoord3f(innerRadiusValue*math.sin(rad), 0.0, innerRadiusValue*math.cos(rad))
                glFT.glVertex3f(innerRadiusValue*math.sin(rad), 0.0, innerRadiusValue*math.cos(rad))
                #glFT.glVertex3f(innerRadiusValue*math.sin(radPlus), 0.0, innerRadiusValue*math.cos(radPlus))

        glFT.glEnd()

        # reset open gl colours
        kLeadColor = 18 # green
        kActiveColor = 15 # white
        kActiveAffectedColor = 8  # purple
        kDormantColor = 4  # blue
        kHiliteColor = 17 # pale blue
        
        glFT.glLineWidth(1.0)
		
        if status == omUI.M3dView.kDormant:
            glFT.glColor3f(colour[0], colour[1], colour[2])

        if (status == omUI.M3dView.kLead):
            view.setDrawColor( kLeadColor, omUI.M3dView.kActiveColors )

        elif (status == omUI.M3dView.kActive):
            view.setDrawColor( kActiveColor, omUI.M3dView.kActiveColors )

        elif (status == omUI.M3dView.kActiveAffected):
            view.setDrawColor( kActiveAffectedColor, omUI.M3dView.kActiveColors )

        elif (status == omUI.M3dView.kDormant):
            view.setDrawColor( kDormantColor, omUI.M3dView.kActiveColors )

        elif (status == omUI.M3dView.kHilite):
            view.setDrawColor( kHiliteColor, omUI.M3dView.kActiveColors )

		# end openGL
        view.endGL()

# creator
def nodeCreator():
    return omMPx.asMPxPtr( mashFalloff() )

# initialiser
def nodeInitializer():
    
    nAttr = om.MFnNumericAttribute()
    mashFalloff.useFalloff = nAttr.create("useFalloff", "uF", om.MFnNumericData.kBoolean,1)
    mashFalloff.addAttribute( mashFalloff.useFalloff )
    
    nAttr = om.MFnNumericAttribute()
    mashFalloff.invertFalloff = nAttr.create("invertFalloff", "inF", om.MFnNumericData.kBoolean,0)
    mashFalloff.addAttribute( mashFalloff.invertFalloff )
    
    nAttr = om.MFnNumericAttribute()
    mashFalloff.aRadius = nAttr.create("radius", "rad", om.MFnNumericData.kFloat, 10)
    nAttr.setMin(0)
    nAttr.setSoftMax(50)
    nAttr.setKeyable(1)
    mashFalloff.addAttribute( mashFalloff.aRadius )
    
    nAttr = om.MFnNumericAttribute()
    mashFalloff.aInnerRadius = nAttr.create("innerRadius", "inRad", om.MFnNumericData.kFloat, 0.3)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashFalloff.addAttribute( mashFalloff.aInnerRadius )
    
    # numeric attr 
    nAttr = om.MFnNumericAttribute()
    mashFalloff.aColor = nAttr.createColor("shapeColor", "sc")
    nAttr.setDefault(0.7, 0.0, 0.35)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    nAttr.setUsedAsColor(1)
    nAttr.setReadable(1)
    nAttr.setWritable(1)
    mashFalloff.addAttribute(mashFalloff.aColor)

    mAttr = om.MFnMessageAttribute()
    mashFalloff.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashFalloff.addAttribute( mashFalloff.falloffMessage )
	
# initialise the script plug-in
def initializePlugin(mobject):
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    try:
        mplugin.registerNode( nodeType, nodeId, nodeCreator, nodeInitializer, omMPx.MPxNode.kLocatorNode )
    except:
        sys.stderr.write( "Failed to register node: %s" % nodeType )
        raise

# uninitialize the script plug-in
def uninitializePlugin(mobject):
    mplugin = omMPx.MFnPlugin(mobject)
    try:
        mplugin.deregisterNode( nodeId )
    except:
        sys.stderr.write( "Failed to deregister node: %s" % nodeType )
        raise
	
