# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_ID"

mashIdId = om.MTypeId(0x0011BE39)

class mashId( omMPx.MPxNode ):
    
    inputs = om.MObject()
    EnvelopeValue = om.MObject()
    outputArray = om.MObject()
    
    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_IDTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_ID" "ID";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Mute Node" -collapse 0;
        
        editorTemplate  -label "ID Type" -addControl "idtype";
        editorTemplate  -label "Num. of Instancer objects" -addControl "numObjects";
        editorTemplate  -label "Random Seed" -addControl "seed";
        //editorTemplate  -label "Group Size" -addControl "grpsize";
        editorTemplate  -label "Randomise Cycle Start Points" -addControl "randomCycle";
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate  -label "Time" -addControl "time";
        
        editorTemplate -suppress "inIterations";
        editorTemplate -suppress "grpsize";

		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):

        if plug != mashId.outputArray:
            return om.kUnknownParameter
        
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(self.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(self.outputArray)

        numberOfCalcs = dataBlock.inputValue(self.inIterations).asInt()
        numberOfCalcsPerm = dataBlock.inputValue(self.inIterations).asInt()
        numberOfObjects = dataBlock.inputValue(self.aNumObjects).asInt()
        seedValue = dataBlock.inputValue(self.aSeed).asInt()
        groupSize = dataBlock.inputValue(self.aGrpSize).asInt()
        idType = dataBlock.inputValue(self.aIdType).asInt()
        randomCycleFlag = dataBlock.inputValue(self.aRandomCycleStart).asInt()
    
        dataType = om.MFnData.kVectorArray
        #Organise the input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #Get the time
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        time = currentTime.value()
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outputArray
        outDData = dataBlock.outputValue(mashId.outputArray)
        nData = om.MFnVectorArrayData()

        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
    
        #print "number of calcs", numberOfCalcs
        
        #we start at 1, because that's the minimum interface value (though this means we have to -1 below)... Hey, it works for me...
        idCounter = 1
        groupCounter = 1
        #resultX = 0
        for i in range(numberOfCalcs):
            if (inDArray.length() > 0):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
            
            #linear
            if idType == 1:
                resultX = (idCounter-1)
                if idCounter < numberOfObjects:
                    idCounter += 1
                else:
                    idCounter = 1
            
            #random
            if idType == 2:
                resultX = random.uniform(0,(numberOfObjects-1))

            #echo
            if idType == 3:
                resultX = (groupCounter-1)
                #if we're at the end of a group, move to the next group, and reset the group i
                if idCounter == numberOfCalcsPerm:
                    groupCounter += 1
                    idCounter = 0
                #if we're in the last group, go back to the beginning.
                if (groupCounter-1) == numberOfObjects:
                    groupCounter = 1
                
                idCounter +=1
    
            #cycle
            if idType == 4:
                if randomCycleFlag == 1:
                    random.seed(i)
                    startPoint = random.randint(0,(numberOfObjects-1))
                    currentResult = startPoint + time
                
                else:
                    currentResult = 0+time
                        
                resultX = int(currentResult%numberOfObjects)

                
            #so as not to break MASH we still output a vector array, but not to worry, SOuPs ATDA node will convert it to a doubleArray for us.
            resultY = resultX
            resultZ = resultX
            '''
            if (inDArray.length() > 0):
                resultX += testRead.x
                resultY += testRead.y
                resultZ += testRead.z
                #
            else:
                resultX = 0
                resultY = 0
                resultZ = 0
            '''
               
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)


        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashId() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashId.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashId.addAttribute( mashId.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashId.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    mashId.addAttribute ( mashId.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashId.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashId.addAttribute( mashId.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashId.aSeed = nAttr.create("seed", "see", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashId.addAttribute( mashId.aSeed )
    mashId.attributeAffects( mashId.aSeed, mashId.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashId.aGrpSize = nAttr.create("grpsize", "gsz", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashId.addAttribute( mashId.aGrpSize )
    mashId.attributeAffects( mashId.aGrpSize, mashId.outputArray)

    nAttr = om.MFnNumericAttribute()
    mashId.aNumObjects = nAttr.create("numObjects", "nuob", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(30)
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashId.addAttribute( mashId.aNumObjects )
    mashId.attributeAffects( mashId.aNumObjects, mashId.outputArray)

    nAttr = om.MFnNumericAttribute()
    mashId.aRandomCycleStart = nAttr.create("randomCycle", "ranCyc", om.MFnNumericData.kBoolean,1)
    mashId.addAttribute( mashId.aRandomCycleStart )
    mashId.attributeAffects( mashId.aRandomCycleStart, mashId.outputArray)
    
    uAttr = om.MFnUnitAttribute()
    mashId.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashId.addAttribute(mashId.aTime)
    mashId.attributeAffects(mashId.aTime, mashId.outputArray)

    typedAttr = om.MFnTypedAttribute()
    enumAttr = om.MFnEnumAttribute()
    mashId.aIdType = enumAttr.create("idtype", "idt", om.MFnData.kNumeric)
    enumAttr.addField('Linear', 1)
    enumAttr.addField('Random', 2)
    enumAttr.addField('Echo ID', 3)
    enumAttr.addField('Cycle', 4)
    #enumAttr.addField('Cycle Group', 5)
    mashId.randomeData = typedAttr.create("idTypeData", "idtd", om.MFnData.kString)
    mashId.addAttribute ( mashId.aIdType )
    mashId.attributeAffects( mashId.aIdType, mashId.outputArray)



    mashId.attributeAffects( mashId.inputArray, mashId.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashIdId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashIdId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    