import maya.OpenMaya as om
import maya.OpenMayaMPx as OpenMayaMPx
import math
import random
import maya.mel as mel
import sys

kPluginNodeTypeName = "MASH_Inherit"
kPluginNodeId = om.MTypeId(0x0011BE11)

class mashInherit(OpenMayaMPx.MPxNode):
    inputs = om.MObject()
    outputArray = om.MObject()
    inTime = om.MObject()

    #AETemplate with radio buttons, swish.
    mel.eval('''
        global proc AEMASH_InheritTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Inherit" "Inherit";

        editorTemplate -beginScrollLayout;
        editorTemplate -beginLayout "Inherit Node" -collapse 0;
        
        editorTemplate -callCustom "AEDistributeRadios"
        "AEDistributeRadiosEdit"
        "inheritStyle";
        
        editorTemplate  -label "Random Seed" -addControl "randomSeed";
        editorTemplate -addSeparator;
        
        editorTemplate  -label "Strength" -addControl "Envelope";
        editorTemplate  -label "Random Strength" -addControl "randEnvelope";
        editorTemplate  -label "Step Strength" -addControl "StepEnvelope";
        editorTemplate -suppress "inIterations";

        editorTemplate -addSeparator;
        
        editorTemplate  -label "Force" -addControl "force";
        editorTemplate  -label "Force Variation" -addControl "forceVar";
        
        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate  -label "Time" -addControl "time";
        
        
        AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        global proc AEDistributeRadios( string $attrName )
        {
       
        radioButtonGrp
		-label "Distribution Style: "
		-numberOfRadioButtons 2
		-label1 "Linear"
		-data1 1
		-label2 "Random"
		-data2 2

		inheritDistRadioGrp;
        connectControl inheritDistRadioGrp $attrName;
        
        }
        
        global proc AEDistributeRadiosEdit( string $attrName )
        {
        
        connectControl inheritDistRadioGrp $attrName;
        
        }
                
        
        ''')

    def __init__(self):
        
        OpenMayaMPx.MPxNode.__init__(self)
        
		
    def compute(self, plug, dataBlock):

        if plug != mashInherit.outputArray:
            return om.kUnknownParameter

        #Organise the input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #get values from the interface
        dataHandle = dataBlock.outputArrayValue(mashInherit.inputs)
        numElements = dataHandle.elementCount()
        currentTime = dataBlock.inputValue(self.inTime).asTime()
        typeOfDistribution = dataBlock.inputValue(mashInherit.inheritStyle).asInt()
        randomSeedValue = dataBlock.inputValue(mashInherit.randomSeed).asInt()
        EnvelopeValue = dataBlock.inputValue(self.aEnvelope).asFloat()
        EnvelopeRandValue = dataBlock.inputValue(self.aRandEnvelope).asFloat()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
        forceValue = dataBlock.inputValue(self.force).asFloat()
        forceVarValue = dataBlock.inputValue(self.forceVar).asFloat()
        
        #organise the out array
        outDData = dataBlock.outputValue(mashInherit.outputArray)
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        numberOfCalcs = dataBlock.inputValue(mashInherit.inIterations).asInt()
        nData = om.MFnVectorArrayData()
        
        #get the falloff object position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
            
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        if inputs.length() > 0:
            
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
            
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
    
        for i in range(numberOfCalcs):
            testRead = om.MVector(0.0,0.0,0.0)
            if (inDArray.length() > 0):
                #if there's an in array, get the appropriate input value
                testRead = inDArray[i]

            random.seed(i*randomSeedValue)
        
            if (typeOfDistribution == 1):
                if (i < numElements):
                    dataHandle.jumpToElement(i)
                    handle = dataHandle.inputValue()
                    result = handle.asFloat()
                    valueComingIn = handle.asFloat3()

                else:
                    valueComingIn = [0,0,0]

            elif (typeOfDistribution == 2):
                desination = random.randint(0,(numElements-1))
                dataHandle.jumpToElement(desination)
                handle = dataHandle.inputValue()
                result = handle.asFloat()
                valueComingIn = handle.asFloat3()
        
            
            randStepValue = (EnvelopeRandValue) * (EnvelopeRandValue) * (3 - 2 * (EnvelopeRandValue))
                
            #randomise the smoothstep so that each object starts and ends at a different time.
            for x in range(0,20):
                whichSmoother = random.randint(0,20)
                if (whichSmoother == 0):
                    #finishes earlier
                    randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                if (whichSmoother == 1):
                    #starts later
                    randStepValue = randStepValue * randStepValue;
            
            A = 1;
            B = 0;
            
            #finish off the smoothstep
            newRandStepValue = (A * randStepValue) + (B * (1 - randStepValue))
            
            stepResult = 1
            if (stepEnvelopeValue < 1):
                #find the scale range for this i
                step = (1.0/(numberOfCalcs+1))*(i+1)
                minStep = (1.0/(numberOfCalcs+1))*i
                
                #are you step enviloping?
                if stepEnvelopeValue <= step:
                    #normalise to get the correct scale value
                    stepResult = (stepEnvelopeValue-minStep)/(step-minStep)
                else:
                    stepResult = 1.0
                
                if stepResult < step:
                    stepResult = 0.0

            resultX = valueComingIn[0]
            resultY = valueComingIn[1]
            resultZ = valueComingIn[2]

            #calculate force values based on the sliders, no need for matrix calcs as it's a waste of slow API calls.
            if (forceValue != 0):
                if (testRead.x == 0):
                    testRead.x = random.uniform(-10, 10)
                if (testRead.y == 0):
                    testRead.y = random.uniform(-10, 10)
                if (testRead.z == 0):
                    testRead.z = random.uniform(-10, 10)
                
                variationX = 1-random.uniform(0,forceVarValue)
                variationY = 1-random.uniform(0,forceVarValue)
                variationZ = 1-random.uniform(0,forceVarValue)

                forceX = testRead.x*(forceValue*variationX)
                forceY = testRead.y*(forceValue*variationY)
                forceZ = testRead.z*(forceValue*variationZ)

                resultX += forceX
                resultY += forceY
                resultZ += forceZ
                
            resultX = (resultX*EnvelopeValue)*newRandStepValue*stepResult
            resultY = (resultY*EnvelopeValue)*newRandStepValue*stepResult
            resultZ = (resultZ*EnvelopeValue)*newRandStepValue*stepResult

            if useFalloffFlag==1:
                if inputs.length() > 0:
                    if radius == 0:
                        radius = 0.001 #avoid divide by zero error
                    
                    #how far is our object from the sphere centre?
                    #fast approximate method, not really good enough.
                    '''
                    distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                    distanceSqrd *= 0.1
                    '''
                    #slow more accurate method
                    xd = testRead.x - falloffObjPos.x
                    yd = testRead.y - falloffObjPos.y
                    zd = testRead.z - falloffObjPos.z
                    distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                    
                    #linear falloff
                    falloff = min(1, max(0, distanceSqrd/radius))
                    
                    #normalise the result to take into account the minimum radius
                    if innerRadius > 0: #avoid divide by zero error
                        falloff = (falloff - innerRadius) / (1-innerRadius)
                    
                    #correct for times when we get a negative result (with an innerZone close to 1)
                    if falloff < 0:
                        falloff = 0
                    
                    #smooth step the linear falloff to make smooth falloff
                    A = 1;
                    B = 0;
                    
                    #trying to decide which of these smooth steps I prefer...
                    #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                    randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                    falloff = (A * randStepValue) + (B * (1 - randStepValue))
                    
                    if invertFalloffFlag !=1:
                        falloffEnvelope = 1-falloff
                    else:
                        falloffEnvelope = falloff
                    
                    resultX *= falloffEnvelope
                    resultY *= falloffEnvelope
                    resultZ *= falloffEnvelope

            
            if (inDArray.length() > 0):
                resultX +=  testRead.x 
                resultY +=  testRead.y 
                resultZ +=  testRead.z 
            
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)
        
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        
        
        outDData.setClean()
        dataBlock.setClean(plug)


def nodeCreator():

    return OpenMayaMPx.asMPxPtr( mashInherit() )
	
def nodeInitializer():
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashInherit.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashInherit.addAttribute( mashInherit.inputArray)
        
    tAttr = om.MFnTypedAttribute()
    mashInherit.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(0)
    tAttr.setKeyable(0)
    tAttr.setReadable(1)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(1)
    mashInherit.addAttribute( mashInherit.outputArray )

    nAttr = om.MFnNumericAttribute()
    mashInherit.inputs = nAttr.create ( "inputs", "in", om.MFnNumericData.k3Float, 1 )
    nAttr.setArray(1)
    nAttr.setWritable(1)
    nAttr.setReadable(1)
    nAttr.setStorable(1)
    mashInherit.addAttribute ( mashInherit.inputs )
    
    uAttr = om.MFnUnitAttribute()
    mashInherit.inTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashInherit.addAttribute(mashInherit.inTime)
    mashInherit.attributeAffects(mashInherit.inTime, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    mashInherit.addAttribute ( mashInherit.inIterations )
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.inheritStyle = nAttr.create("inheritStyle", "inSty", om.MFnNumericData.kInt, 1)
    nAttr.setStorable(1)
    mashInherit.addAttribute( mashInherit.inheritStyle )
    mashInherit.attributeAffects( mashInherit.inheritStyle, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.randomSeed = nAttr.create("randomSeed", "raSe", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setStorable(1)
    mashInherit.addAttribute( mashInherit.randomSeed )
    mashInherit.attributeAffects( mashInherit.randomSeed, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.aEnvelope = nAttr.create("Envelope", "env", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashInherit.addAttribute( mashInherit.aEnvelope )
    mashInherit.attributeAffects( mashInherit.aEnvelope, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.aRandEnvelope = nAttr.create("randEnvelope", "raEn", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashInherit.addAttribute( mashInherit.aRandEnvelope )
    mashInherit.attributeAffects( mashInherit.aRandEnvelope, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashInherit.addAttribute( mashInherit.aStepEnvelope )
    mashInherit.attributeAffects( mashInherit.aStepEnvelope, mashInherit.outputArray)
    
    mashInherit.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashInherit.addAttribute(mashInherit.falloffObject)
    mashInherit.attributeAffects(mashInherit.falloffObject, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.force = nAttr.create ( "force", "for", om.MFnNumericData.kFloat, 0 )
    nAttr.setHidden(0)
    nAttr.setSoftMin(-2)
    nAttr.setSoftMax(2)
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashInherit.addAttribute ( mashInherit.force )
    mashInherit.attributeAffects(mashInherit.force, mashInherit.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashInherit.forceVar = nAttr.create ( "forceVar", "forV", om.MFnNumericData.kFloat, 0 )
    nAttr.setHidden(0)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    mashInherit.addAttribute ( mashInherit.forceVar )
    mashInherit.attributeAffects(mashInherit.forceVar, mashInherit.outputArray)
    
    mAttr = om.MFnMessageAttribute()
    mashInherit.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashInherit.addAttribute( mashInherit.falloffMessage )

    mashInherit.attributeAffects( mashInherit.inputs, mashInherit.outputArray)


def initializePlugin(mobject):
    fnPlugin = OpenMayaMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    try:
        fnPlugin.registerNode(kPluginNodeTypeName, kPluginNodeId, nodeCreator, nodeInitializer, OpenMayaMPx.MPxNode.kDependNode)
    except:
        sys.stderr.write( "Failed to register node: %s\n" % kPluginNodeTypeName )


def uninitializePlugin(mobject):
    fnPlugin = OpenMayaMPx.MFnPlugin(mobject)
    try:
        fnPlugin.deregisterNode(kPluginNodeId)
    except:
        sys.stderr.write( "Failed to unregister node: %s\n" % kPluginNodeTypeName )