# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Mute"

mashMuteId = om.MTypeId(0x0011BE28)

class mashMute( omMPx.MPxNode ):
    
    inputs = om.MObject()
    EnvelopeValue = om.MObject()
    outputArray = om.MObject()
    
    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_MuteTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Mute" "Mute";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Mute Node" -collapse 0;
        
        editorTemplate  -label "Strength" -addControl "Envelope";
        editorTemplate  -label "Random Strength" -addControl "RandomEnvelope";
        editorTemplate  -label "Step Strength" -addControl "StepEnvelope";
        
        editorTemplate -suppress "inIterations";
        

		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):

        if plug != mashMute.outputArray:
            return om.kUnknownParameter
        
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashMute.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashMute.outputArray)

        numberOfCalcs = dataBlock.inputValue(mashMute.inIterations).asInt()
        EnvelopeValue = dataBlock.inputValue(mashMute.aEnvelope).asFloat()
        randEnvelopeValue = dataBlock.inputValue(mashMute.aRandEnvelope).asFloat()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
    
        dataType = om.MFnData.kVectorArray
        #Organise the input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outputArray
        outDData = dataBlock.outputValue(mashMute.outputArray)
        nData = om.MFnVectorArrayData()

        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()

        for i in range(numberOfCalcs):
            if (inDArray.length() > 0):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
            
            # Get the element index
            index = i
            random.seed(index)
            inNormalArray = om.MFloatArray()
            inNormalArray.clear()
            
            #find the scale range for this i
            step = (1.0/(numberOfCalcs+1))*(i+1)
            min = (1.0/(numberOfCalcs+1))*i
    
            #if i  < 2:
            #    print "max", step, "min", min, "slider", stepEnvelopeValue
    
            #are you step enviloping?
            if stepEnvelopeValue <= step:
                #normalise to get the correct scale value
                stepResult = (stepEnvelopeValue-min)/(step-min)
            else:
                stepResult = 1.0
                
            if stepResult < step:
                stepResult = 0.0
    
            #smoothDistribute the randEnvelopeValue
            #using the old school smoothDistribute calculation.
            if (randEnvelopeValue < 1):
                randStepValue = (randEnvelopeValue) * (randEnvelopeValue) * (3 - 2 * (randEnvelopeValue))
                
                #randomise the smoothstep so that each object starts and ends at a different time.
                for x in range(0,20):
                    whichSmoother = random.randint(0,20)
                    if (whichSmoother == 0):
                        #finishes earlier
                        randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                    if (whichSmoother == 1):
                        #starts later
                        randStepValue = randStepValue * randStepValue;
                
                A = 1;
                B = 0;
                
                #finish off the smoothstep
                newrandEnvelopeValue = (A * randStepValue) + (B * (1 - randStepValue))
            else:
                newrandEnvelopeValue = randEnvelopeValue
            
            if (inDArray.length() > 0):
                resultX = (testRead.x*(1.0-EnvelopeValue))*(1.0-newrandEnvelopeValue)*(1.0-stepResult)
                resultY = (testRead.y*(1.0-EnvelopeValue))*(1.0-newrandEnvelopeValue)*(1.0-stepResult)
                resultZ = (testRead.z*(1.0-EnvelopeValue))*(1.0-newrandEnvelopeValue)*(1.0-stepResult)
                #
            else:
                resultX = 0
                resultY = 0
                resultZ = 0

               
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)


        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashMute() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashMute.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashMute.addAttribute( mashMute.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashMute.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    mashMute.addAttribute ( mashMute.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashMute.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashMute.addAttribute( mashMute.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashMute.aEnvelope = nAttr.create("Envelope", "env", om.MFnNumericData.kFloat, 0)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashMute.addAttribute( mashMute.aEnvelope )
    mashMute.attributeAffects( mashMute.aEnvelope, mashMute.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashMute.aRandEnvelope = nAttr.create("RandomEnvelope", "RaEnv", om.MFnNumericData.kFloat, 0)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashMute.addAttribute( mashMute.aRandEnvelope )
    mashMute.attributeAffects( mashMute.aRandEnvelope, mashMute.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashMute.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 0)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashMute.addAttribute( mashMute.aStepEnvelope )
    mashMute.attributeAffects( mashMute.aStepEnvelope, mashMute.outputArray)

    mashMute.attributeAffects( mashMute.inputArray, mashMute.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashMuteId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashMuteId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    