# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Noise"

mashNoiseId = om.MTypeId(0x0011BEE1)

class mashNoise( omMPx.MPxNode ):
    
    time = om.MObject()
    inputArray   = om.MObject()
    outputArray   = om.MObject()
    Envelope = om.MObject()
    frequencyZ = om.MObject()
    amplitudeZ = om.MObject()
    frequencyY = om.MObject()
    amplitudeY = om.MObject()
    frequencyX = om.MObject()
    amplitudeX = om.MObject()
    modular = om.MObject()
    step = om.MObject()
    offsetValues = om.MObject()

    
    #AETemplate in the .py, ahhh yeahhh.
    mel.eval('''
        global proc AEMASH_NoiseTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Noise" "Noise";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Noise Node" -collapse 0;
        
        editorTemplate  -label "Amplitude X" -addControl "amplitudeX";
        editorTemplate  -label "Amplitude Y" -addControl "amplitudeY";
        editorTemplate  -label "Amplitude Z" -addControl "amplitudeZ";
        editorTemplate -addSeparator;
        editorTemplate  -label "Frequency X" -addControl "frequencyX";
        editorTemplate  -label "Frequency Y" -addControl "frequencyY";
        editorTemplate  -label "Frequency Z" -addControl "frequencyZ";

        editorTemplate -addSeparator;
        editorTemplate  -label "Strength" -addControl "Envelope";
        editorTemplate  -label "Random Strength" -addControl "randEnvelope";
        editorTemplate  -label "Step Strength" -addControl "StepEnvelope";
        
        editorTemplate -addSeparator;
        editorTemplate  -label "Step" -addControl "step";
        editorTemplate  -label "Pattern" -addControl "modular";
        editorTemplate  -label "Octaves" -addControl "octaves";
        editorTemplate  -label "Random Seed" -addControl "seed";
        editorTemplate  -label "Absolute Noise" -addControl "positiveNoise";
        editorTemplate -suppress "inIterations";
        
        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate  -label "Time" -addControl "time";
        editorTemplate  -label "Offset Values" -addControl "offsetValues";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        
        global proc AEfalloffButton ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Create Falloff Object"  -c ("falloffButtonCMDS " + $nodeName[0] + " 1")
        setFalloffMASHButton;
        
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        global proc AEfalloffButtonEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("falloffButtonCMDS " + $nodeName[0] + " 1")
        setFalloffMASHButton;
        
        }
        
        
        global proc falloffButtonCMDS (string $nodeName, int $whichCMD) {
        if ($whichCMD == 1) {
        //string $inRotation[] = `listConnections -d 0 -s 1 ($nodeName+".inRotationPP")`;
        
        string $falloffNode = `createNode MASH_Falloff`;
        
        string $transform = "";
        string $parents[] = `listRelatives -fullPath -parent $falloffNode`;
        $transform = $parents[0];
        
        connectAttr -force ($nodeName+".falloffMessage") ($falloffNode+".falloffMessage") ;   
        connectAttr -force ($transform+".translate") ($nodeName+".falloffObject");        
        }

        evalDeferred("updateAE " + $nodeName);
        }
        
        
        ''')

    
    def __init__( self ):
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):
            
        
        if plug != mashNoise.outputArray:
            return om.kUnknownParameter
        
        # get the input handle
        #dataHandle = dataBlock.inputValue(mashNoise.time).asFloat()
    
        currentTime = dataBlock.inputValue(self.time).asTime()
        dataHandle = currentTime.value()
        
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
    
        # get the interface values
        amplitudeValueX = dataBlock.inputValue(mashNoise.aAmplitudeX).asFloat()
        frequencyValueX = dataBlock.inputValue(mashNoise.aFrequencyX).asFloat()
        amplitudeValueY = dataBlock.inputValue(mashNoise.aAmplitudeY).asFloat()
        frequencyValueY = dataBlock.inputValue(mashNoise.aFrequencyY).asFloat()
        amplitudeValueZ = dataBlock.inputValue(mashNoise.aAmplitudeZ).asFloat()
        frequencyValueZ = dataBlock.inputValue(mashNoise.aFrequencyZ).asFloat()
        stepValue = dataBlock.inputValue(mashNoise.aStep).asFloat()
        modularValue = dataBlock.inputValue(mashNoise.aModular).asInt()
        octavesValue = dataBlock.inputValue(mashNoise.aOctaves).asInt()
        numberOfCalcs = dataBlock.inputValue(mashNoise.inIterations).asInt()
        seedValue = dataBlock.inputValue(mashNoise.aSeed).asInt()
        positiveFlag = dataBlock.inputValue(mashNoise.aPositive).asBool()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
        EnvelopeValue = dataBlock.inputValue(mashNoise.aEnvelope).asFloat()
        EnvelopeRandValue = dataBlock.inputValue(mashNoise.aRandEnvelope).asFloat()
        
        offsetValuesArray = dataBlock.inputValue(mashNoise.offsetValues).asFloat3()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashNoise.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashNoise.outputArray)

        #create the output array
        outResultArray = om.MVectorArray()
        outResultArray.clear()

        inDArray = om.MVectorArray()
        inDArray.clear()
        
        #Organise the input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #Set the outputArray
        outDData = dataBlock.outputValue(mashNoise.outputArray)
        nData = om.MFnVectorArrayData()
        
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        if inputs.length() > 0:
                        
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
        
        if (dataHandle == 0):
            dataHandle = 1
                
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
            
        #make some noise!
        for i in range(numberOfCalcs):
            #if there's an in array, get those values.
            if (inDArray.length() > 0):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
            else:
                testRead = om.MVector(0.0,0.0,0.0)

            
            #write an extra line of code for no reason at all.
            index = i
            stepAddition = stepValue*index
            
            #clear results
            resultX = 0
            resultY = 0
            resultZ = 0
            
            #set seed for smoothstepping
            random.seed(index*seedValue)
            
            #refresh the amp and freq values for the next calculation
            amplitudeValueX = dataBlock.inputValue(mashNoise.aAmplitudeX).asFloat()
            frequencyValueX = dataBlock.inputValue(mashNoise.aFrequencyX).asFloat()
            amplitudeValueY = dataBlock.inputValue(mashNoise.aAmplitudeY).asFloat()
            frequencyValueY = dataBlock.inputValue(mashNoise.aFrequencyY).asFloat()
            amplitudeValueZ = dataBlock.inputValue(mashNoise.aAmplitudeZ).asFloat()
            frequencyValueZ = dataBlock.inputValue(mashNoise.aFrequencyZ).asFloat()
        
            #calculate the noise, note the octives
            if (i%modularValue == 0):
                for _ in range(octavesValue):
                    
                    resultX += (improvedGradNoise( (dataHandle+stepAddition)*(frequencyValueX*0.0101)+(100*seedValue), (dataHandle+stepAddition)*(frequencyValueX*0.0101)+(100*seedValue), (dataHandle+stepAddition)*(frequencyValueX*0.0101)+(100*seedValue) )) * amplitudeValueX
            
                    resultY += (improvedGradNoise( (dataHandle+stepAddition)*(frequencyValueY*0.0101)+(250*seedValue), (dataHandle+stepAddition)*(frequencyValueY*0.0101)+(250*seedValue), (dataHandle+stepAddition)*(frequencyValueY*0.0101)+(250*seedValue) )) * amplitudeValueY
       
                    resultZ += (improvedGradNoise( (dataHandle+stepAddition)*(frequencyValueZ*0.0101)+(500*seedValue), (dataHandle+stepAddition)*(frequencyValueZ*0.0101)+500*seedValue, (dataHandle+stepAddition)*(frequencyValueZ*0.0101)+(500*seedValue) )) * amplitudeValueZ
              
                    frequencyValueX *=2
                    frequencyValueY *=2
                    frequencyValueZ *=2
                    amplitudeValueX /= 2
                    amplitudeValueY /= 2
                    amplitudeValueZ /= 2
            
                #If the positive flag is set, clamp the values below zero - I'm playing with how I'd like this to work, currently it just effects Y (useful for scale). Independent controls perhaps?
                
                if (positiveFlag!=0):
                    if (resultX < 0):
                        resultX = math.fabs(resultX)
                    if (resultY < 0):
                        resultY = math.fabs(resultY)
                    if (resultZ < 0):
                        resultZ = math.fabs(resultZ)
            
                #if there's an in array, add those values to the results.
            
                
                if (EnvelopeRandValue < 1):
                    randStepValue = (EnvelopeRandValue) * (EnvelopeRandValue) * (3 - 2 * (EnvelopeRandValue))
                    
                    #randomise the smoothstep so that each object starts and ends at a different time.
                    for x in range(0,20):
                        whichSmoother = random.randint(0,20)
                        if (whichSmoother == 0):
                            #finishes earlier
                            randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                        if (whichSmoother == 1):
                            #starts later
                            randStepValue = randStepValue * randStepValue;
                    
                    A = 1;
                    B = 0;
                    
                    #finish off the smoothstep
                    newRandStepValue = (A * randStepValue) + (B * (1 - randStepValue))
                else:
                    newRandStepValue = EnvelopeRandValue
            
                stepResult = 1
                if (stepEnvelopeValue < 1):
                    #find the scale range for this i
                    step = (1.0/(numberOfCalcs+1))*(i+1)
                    minStep = (1.0/(numberOfCalcs+1))*i
                    
                    #are you step enviloping?
                    if stepEnvelopeValue <= step:
                        #normalise to get the correct scale value
                        stepResult = (stepEnvelopeValue-minStep)/(step-minStep)
                    else:
                        stepResult = 1.0
                    
                    if stepResult < step:
                        stepResult = 0.0
                
                resultX = offsetValuesArray[0]+((resultX * EnvelopeValue)*newRandStepValue)*stepResult
                resultY = offsetValuesArray[1]+((resultY * EnvelopeValue)*newRandStepValue)*stepResult
                resultZ = offsetValuesArray[2]+((resultZ * EnvelopeValue)*newRandStepValue)*stepResult
                
                if useFalloffFlag==1:
                    if inputs.length() > 0:
                        if radius == 0:
                            radius = 0.001 #avoid divide by zero error
                        
                        #how far is our object from the sphere centre?
                        #fast approximate method, not really good enough.
                        '''
                        distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                        distanceSqrd *= 0.1
                        '''
                        #slow more accurate method
                        xd = testRead.x - falloffObjPos.x
                        yd = testRead.y - falloffObjPos.y
                        zd = testRead.z - falloffObjPos.z
                        distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                        
                        #linear falloff
                        falloff = min(1, max(0, distanceSqrd/radius))
                        
                        #normalise the result to take into account the minimum radius
                        if innerRadius > 0: #avoid divide by zero error
                            falloff = (falloff - innerRadius) / (1-innerRadius)
                        
                        #correct for times when we get a negative result (with an innerZone close to 1)
                        if falloff < 0:
                            falloff = 0
                        
                        #smooth step the linear falloff to make smooth falloff
                        A = 1;
                        B = 0;
                        
                        #trying to decide which of these smooth steps I prefer...
                        #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                        randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                        falloff = (A * randStepValue) + (B * (1 - randStepValue))
                        
                        if invertFalloffFlag !=1:
                            falloffEnvelope = 1-falloff
                        else:
                            falloffEnvelope = falloff
                                           
                        resultX *= falloffEnvelope
                        resultY *= falloffEnvelope
                        resultZ *= falloffEnvelope
            
            if (inDArray.length() > 0):
                resultX += testRead.x
                resultY += testRead.y 
                resultZ += testRead.z
            
            #print "noise results after addn ", resultX, resultY, resultZ
            
            #set this result
            thisResult = om.MVector(resultX,resultY,resultZ)
            
            #add the result to the out array
            outResultArray.append(thisResult)

        #set the output, or ahhtput if you're from London.
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        
        #do some gentle dusting of the dataBlock
        dataBlock.setClean(plug)

def nodeCreator():
    
    return omMPx.asMPxPtr( mashNoise() )

def nodeInitializer():
    
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashNoise.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(0)
    tAttr.setReadable(True)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashNoise.addAttribute( mashNoise.inputArray)
    
    tAttr = om.MFnTypedAttribute()
    mashNoise.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(0)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(0)
    tAttr.setUsesArrayDataBuilder(True)
    mashNoise.addAttribute( mashNoise.outputArray )
    
    uAttr = om.MFnUnitAttribute()
    mashNoise.time = uAttr.create ( "time", "ti", om.MFnUnitAttribute.kTime, 1 )
    uAttr.setStorable(1)
    uAttr.setKeyable(1)
    mashNoise.addAttribute ( mashNoise.time )
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashNoise.addAttribute ( mashNoise.inIterations )

    nAttr = om.MFnNumericAttribute()
    mashNoise.offsetValues = nAttr.create( "offsetValues", "ofVal", om.MFnNumericData.k3Float, 0.0 )
    nAttr.setStorable(1)
    nAttr.setWritable(1)
    mashNoise.addAttribute( mashNoise.offsetValues )
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aStep = nAttr.create("step", "st", om.MFnNumericData.kFloat, 200)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(200)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aModular = nAttr.create("modular", "mod", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aAmplitudeX = nAttr.create("amplitudeX", "ampX", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aFrequencyX = nAttr.create("frequencyX", "freqX", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aAmplitudeY = nAttr.create("amplitudeY", "ampY", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aFrequencyY = nAttr.create("frequencyY", "freqY", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aAmplitudeZ = nAttr.create("amplitudeZ", "ampZ", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aFrequencyZ = nAttr.create("frequencyZ", "freqZ", om.MFnNumericData.kFloat, 1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aEnvelope = nAttr.create("Envelope", "env", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashNoise.addAttribute( mashNoise.aEnvelope )
    mashNoise.attributeAffects( mashNoise.aEnvelope, mashNoise.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aRandEnvelope = nAttr.create("randEnvelope", "raEn", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashNoise.addAttribute( mashNoise.aRandEnvelope )
    mashNoise.attributeAffects( mashNoise.aRandEnvelope, mashNoise.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashNoise.addAttribute( mashNoise.aStepEnvelope )
    mashNoise.attributeAffects( mashNoise.aStepEnvelope, mashNoise.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aOctaves = nAttr.create("octaves", "oct", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(5)
    nAttr.setKeyable(1)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aSeed = nAttr.create("seed", "see", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setKeyable(1)
    mashNoise.addAttribute( mashNoise.aSeed )
    
    mashNoise.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashNoise.addAttribute(mashNoise.falloffObject)
    mashNoise.attributeAffects(mashNoise.falloffObject, mashNoise.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashNoise.aPositive = nAttr.create("positiveNoise", "posNoi", om.MFnNumericData.kBoolean,0)
    mashNoise.addAttribute( mashNoise.aPositive )
    
    mashNoise.addAttribute( mashNoise.aOctaves )
    
    mashNoise.addAttribute( mashNoise.aModular )
    
    mAttr = om.MFnMessageAttribute()
    mashNoise.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashNoise.addAttribute( mashNoise.falloffMessage )

    mashNoise.addAttribute( mashNoise.aStep )
    mashNoise.addAttribute( mashNoise.aAmplitudeX )
    mashNoise.addAttribute( mashNoise.aFrequencyX )
    mashNoise.addAttribute( mashNoise.aAmplitudeY )
    mashNoise.addAttribute( mashNoise.aFrequencyY )
    mashNoise.addAttribute( mashNoise.aAmplitudeZ )
    mashNoise.addAttribute( mashNoise.aFrequencyZ )
	
    mashNoise.attributeAffects( mashNoise.inputArray, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aPositive, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aSeed, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aModular, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aOctaves, mashNoise.outputArray)


    mashNoise.attributeAffects( mashNoise.offsetValues, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.time, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aStep, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aAmplitudeX, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aFrequencyX, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aAmplitudeY, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aFrequencyY, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aAmplitudeZ, mashNoise.outputArray)
    mashNoise.attributeAffects( mashNoise.aFrequencyZ, mashNoise.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashNoiseId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashNoiseId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    

# Ken's permutation array composed of 512 elements = 2 sets of (0,255)
p = [151,160,137, 91, 90, 15,131, 13,201, 95, 96, 53,194,233,  7,225,
	 140, 36,103, 30, 69,142,  8, 99, 37,240, 21, 10, 23,190,  6,148,
	 247,120,234, 75,  0, 26,197, 62, 94,252,219,203,117, 35, 11, 32,
     57,177, 33, 88,237,149, 56, 87,174, 20,125,136,171,168, 68,175,
     74,165, 71,134,139, 48, 27,166, 77,146,158,231, 83,111,229,122,
     60,211,133,230,220,105, 92, 41, 55, 46,245, 40,244,102,143, 54,
     65, 25, 63,161,  1,216, 80, 73,209, 76,132,187,208, 89, 18,169,
	 200,196,135,130,116,188,159, 86,164,100,109,198,173,186,  3, 64,
     52,217,226,250,124,123,  5,202, 38,147,118,126,255, 82, 85,212,
	 207,206, 59,227, 47, 16, 58, 17,182,189, 28, 42,223,183,170,213,
	 119,248,152,  2, 44,154,163, 70,221,153,101,155,167, 43,172,  9,
	 129, 22, 39,253, 19, 98,108,110, 79,113,224,232,178,185,112,104,
	 218,246, 97,228,251, 34,242,193,238,210,144, 12,191,179,162,241,
     81, 51,145,235,249, 14,239,107, 49,192,214, 31,181,199,106,157,
	 184, 84,204,176,115,121, 50, 45,127,  4,150,254,138,236,205, 93,
	 222,114, 67, 29, 24, 72,243,141,128,195, 78, 66,215, 61,156,180,
	 151,160,137, 91, 90, 15,131, 13,201, 95, 96, 53,194,233,  7,225,
	 140, 36,103, 30, 69,142,  8, 99, 37,240, 21, 10, 23,190,  6,148,
	 247,120,234, 75,  0, 26,197, 62, 94,252,219,203,117, 35, 11, 32,
     57,177, 33, 88,237,149, 56, 87,174, 20,125,136,171,168, 68,175,
     74,165, 71,134,139, 48, 27,166, 77,146,158,231, 83,111,229,122,
     60,211,133,230,220,105, 92, 41, 55, 46,245, 40,244,102,143, 54,
     65, 25, 63,161,  1,216, 80, 73,209, 76,132,187,208, 89, 18,169,
	 200,196,135,130,116,188,159, 86,164,100,109,198,173,186,  3, 64,
     52,217,226,250,124,123,  5,202, 38,147,118,126,255, 82, 85,212,
	 207,206, 59,227, 47, 16, 58, 17,182,189, 28, 42,223,183,170,213,
	 119,248,152,  2, 44,154,163, 70,221,153,101,155,167, 43,172,  9,
	 129, 22, 39,253, 19, 98,108,110, 79,113,224,232,178,185,112,104,
	 218,246, 97,228,251, 34,242,193,238,210,144, 12,191,179,162,241,
     81, 51,145,235,249, 14,239,107, 49,192,214, 31,181,199,106,157,
	 184, 84,204,176,115,121, 50, 45,127,  4,150,254,138,236,205, 93,
	 222,114, 67, 29, 24, 72,243,141,128,195, 78, 66,215, 61,156,180]

# Ken's Utility functions for noise
# linear interpolation
def lerp(parameter=0.5,value1=0.0,value2=1.0):
	return value1 + parameter * (value2 - value1)

# Ken's new spline interpolation
def fade(parameter=1.0):
	return parameter*parameter*parameter*(parameter*(parameter*6.0 - 15.0) + 10.0)

# Ken's new function to return gradient values
# based on bit operations on hashId
def grad(hashId=255,x=1.0,y=1.0,z=1.0):
	h = hashId & 15
	if (h < 8):
		u = x
	else:
		u = y
    
	if (h < 4):
		v = y
	elif (h==12 or h==14):
		v = x
	else:
		v = z
	if ((h&1)!=0):
		u = -u
	if ((h&2)!=0):
		v = -v
	return u + v

# Ken's improved gradient noise function
def improvedGradNoise(vx=1.0,vy=1.0,vz=1.0):
	# get integer lattice values for sample point position
	X = int(math.floor(vx)) & 255
	Y = int(math.floor(vy)) & 255
	Z = int(math.floor(vz)) & 255
	# fractional part of point position
	vx -= math.floor(vx)
	vy -= math.floor(vy)
	vz -= math.floor(vz)
	# interpolate fractional part of point position
	u = fade(vx)
	v = fade(vy)
	w = fade(vz)
	# new hash integer lattice cell coords onto perm array
	A = p[X]+Y
	B = p[X+1]+Y
	AA = p[A]+Z
	BA = p[B]+Z
	AB = p[A+1]+Z
	BB = p[B+1]+Z
	# new hash onto gradients
	gradAA  = grad(p[AA],   vx,     vy,     vz  )
	gradBA  = grad(p[BA],   vx-1.0, vy,     vz  )
	gradAB  = grad(p[AB],   vx,     vy-1.0, vz  )
	gradBB  = grad(p[BB],   vx-1.0, vy-1.0, vz  )
	gradAA1 = grad(p[AA+1], vx,     vy,     vz-1.0)
	gradBA1 = grad(p[BA+1], vx-1.0, vy,     vz-1.0)
	gradAB1 = grad(p[AB+1], vx,     vy-1.0, vz-1.0)
	gradBB1 = grad(p[BB+1], vx-1.0, vy-1.0, vz-1.0)
	# trilinear intropolation of resulting gradients to sample point position
	result = lerp(w, lerp(v, lerp(u, gradAA, gradBA), lerp(u, gradAB, gradBB)), lerp(v, lerp(u, gradAA1, gradBA1), lerp(u, gradAB1, gradBB1)))
	return result