# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Orient"

mashRotateId = om.MTypeId(0x87004)

class mashRotate( omMPx.MPxNode ):
    
    inputs = om.MObject()
    outRotationPP = om.MObject()
    
    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_OrientTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Orient" "Orient";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Orient Node" -collapse 0;
        
        editorTemplate  -label "Strength X" -addControl "EnvelopeX";
        editorTemplate  -label "Strength Y" -addControl "EnvelopeY";
        editorTemplate  -label "Strength Z" -addControl "EnvelopeZ";
        
        editorTemplate -beginLayout "Target Mode" -collapse 0;
        editorTemplate  -label "Target Mode" -addControl "targetMode";
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "targetInput" "Target Input";
        editorTemplate -callCustom "AEorientButton" "AEorientButtonEdit" "";
        editorTemplate -endLayout;

        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -suppress "inIterations";
        
        editorTemplate -beginLayout "Inputs" -collapse 1;
        editorTemplate  -label "Time" -addControl "time";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        global proc AEaddSpringButton ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Add Spring Node"  -c ("orientButtonCMDS " + $nodeName[0] + " 1")
        addRotSpringMASHButton;
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        
        global proc AEaddSpringButtonEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("orientButtonCMDS " + $nodeName[0] + " 1")
        addRotSpringMASHButton;
        
        }
        
        global proc orientButtonCMDS (string $nodeName, int $whichCMD)
        {
        if ($whichCMD == 1)
        {
        string $outgoing[] = `listConnections -d 1 -s 0 ($nodeName+".outRotationPP")`;
        
        //create new node and hook up time
        string $nodeType = "MASH_Spring";
        $newNode = `createNode MASH_Spring -n $nodeType`;
        connectAttr -f time1.outTime ($newNode +".time");
        //connect message
        string $conns[] = `listConnections -d 1 ($outgoing[0] +".setMessage")`;
        int $numConns = `size($conns)`;
        int $nextCon = $numConns+1;
        connectAttr -force ($newNode +".message") ($outgoing[0] +".setMessage[" + $nextCon + "]");
        // connect new node to waiter
        connectAttr -force ($newNode +".outArray") ($outgoing[0] +".inRotationPP");
        //connect new node to current node
        connectAttr -force ($nodeName +".outRotationPP") ($newNode +".inputArray");
        //set good rotation defaults
        setAttr ($newNode +".damping") 0.85;
        setAttr ($newNode +".stiffness") 0.05;
        }
        }
        
        global proc AEorientButton ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Set Target"  -c ("orientButtonCMDS " + $nodeName[0] + " 1")
        setOrientMASHButton;
        
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        global proc AEorientButtonEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("orientButtonCMDS " + $nodeName[0] + " 1")
        setOrientMASHButton;
        
        }
        
        
        global proc orientButtonCMDS (string $nodeName, int $whichCMD) {
        if ($whichCMD == 1) {
        string $target[] = `ls -sl -type "transform"`; 
        
        if (size($target) > 0) {
        
        connectAttr -force ($target[0]+".translate") ($nodeName+".targetInput");
        setAttr ($nodeName+".targetMode") 1;
        print "Target connected.";
        
        
        }
        else {
        warning "Please select a translate.";
        }
        }
        evalDeferred("updateAE " + $nodeName);
        }
        
        ''')
    
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
        self._prevLocation = [0,0,0,0]
        self._previousTime = om.MTime()
        self._initialised = False
    
    def compute(self, plug, dataBlock):

        if plug != mashRotate.outRotationPP:
            return om.kUnknownParameter
        
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashRotate.inputArray)
        #Get output array
        outRotationPP_dataHandle = dataBlock.outputValue(mashRotate.outRotationPP)
    
        targetInputValue = dataBlock.inputValue(mashRotate.targetInput).asFloat3()
        
        targetFlag = dataBlock.inputValue(mashRotate.targetMode).asBool()

        numberOfCalcs = dataBlock.inputValue(mashRotate.inIterations).asInt()
        
        envXValue = dataBlock.inputValue(mashRotate.aEnvelopeX).asFloat()
        envYValue = dataBlock.inputValue(mashRotate.aEnvelopeY).asFloat()
        envZValue = dataBlock.inputValue(mashRotate.aEnvelopeZ).asFloat()
        
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        time = currentTime.value()
        
        dataType = om.MFnData.kVectorArray
        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        inRotDArray = om.MVectorArray()
        inRotDArray.clear()
        inRotDArrData = dataBlock.inputValue(self.inRotationPP).data()
        inRotDArrFn = om.MFnVectorArrayData(inRotDArrData)
        inRotDArray = inRotDArrFn.array()
    
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outRotationPP
        outDData = dataBlock.outputValue(mashRotate.outRotationPP)
        nData = om.MFnVectorArrayData()

        #get the falloff object position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
        
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
        
        if inputs.length() > 0:
 
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
        
        timeStep = currentTime.value() - self._previousTime.value()
        if timeStep > 1.0 or timeStep < 0.0:
            self._initialised = False
            self._previousTime = currentTime
            return
        
        if not self._initialised:
            #empty the lists
            del self._prevLocation[:]
            
            for h in range(numberOfCalcs):
                testRead = om.MFloatVector(0.0,0.0,0.0)
                self._prevLocation.append(testRead)
                
            
            self._previousTime = currentTime
            self._initialised = True
        
        
        for i in range(numberOfCalcs):
            if (inDArray.length() > 0):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
    
            if (inRotDArray.length() > 0):
                rotRead = om.MVector(0.0,0.0,0.0)
                rotRead = inRotDArray[i]
            
            resultX = 0
            resultY = 0
            resultZ = 0
            
            inNormalArray = om.MFloatArray()
            inNormalArray.clear()
            
            if (targetFlag !=1):
                #make vectors of the loc positions
                start = om.MVector(self._prevLocation[i])
                end = om.MVector(testRead)
                aimX =  end - start
                #get the delta (direction of travel)
                
            else:
                start = om.MVector(targetInputValue[0],targetInputValue[1],targetInputValue[2])
                end = om.MVector(testRead)
                aimX =  start - end
            
            #tempY because I only have 1 vector to go off
            tempAimY = om.MVector(0,1,0)
            aimZ = aimX^tempAimY
            aimZ.normalize()
            aimY = aimZ^aimX
            aimY.normalize()

            #create a matrix and fill it with the values
            matrix = om.MMatrix()
            om.MScriptUtil.setDoubleArray(matrix[0], 0, aimX.x)
            om.MScriptUtil.setDoubleArray(matrix[0], 1, aimX.y)
            om.MScriptUtil.setDoubleArray(matrix[0], 2, aimX.z)
            om.MScriptUtil.setDoubleArray(matrix[1], 0, aimY.x)
            om.MScriptUtil.setDoubleArray(matrix[1], 1, aimY.y)
            om.MScriptUtil.setDoubleArray(matrix[1], 2, aimY.z)
            om.MScriptUtil.setDoubleArray(matrix[2], 0, aimZ.x)
            om.MScriptUtil.setDoubleArray(matrix[2], 1, aimZ.y)
            om.MScriptUtil.setDoubleArray(matrix[2], 2, aimZ.z)

            #put the matrix into a transform matrix
            mTransformMtx = om.MTransformationMatrix(matrix)

            #get the euler rotations
            eulerRot = mTransformMtx.eulerRotation()
            
            #convert to degrees
            angles = [math.degrees(angle) for angle in (eulerRot.x, eulerRot.y, eulerRot.z)]
            
            if (targetFlag !=1):
                self._prevLocation[i] = om.MVector(testRead.x, testRead.y, testRead.z)
            
            if (inDArray.length() > 0):
                resultX = angles[0]*envXValue
                resultY = angles[1]*envYValue
                resultZ = angles[2]*envZValue
                #
            else:
                resultX = 0
                resultY = 0
                resultZ = 0
    
            #if there's a falloff object and it's flag is on, do the calc
            if useFalloffFlag==1:
                if inputs.length() > 0:
                    if radius == 0:
                        radius = 0.001 #avoid divide by zero error
                    
                    #how far is our object from the sphere centre?
                    #fast approximate method, not really good enough.
                    '''
                    distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                    distanceSqrd *= 0.1
                    '''
                    #slow more accurate method
                    xd = testRead.x - falloffObjPos.x
                    yd = testRead.y - falloffObjPos.y
                    zd = testRead.z - falloffObjPos.z
                    distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                    
                    #linear falloff
                    falloff = min(1, max(0, distanceSqrd/radius))
                    
                    #normalise the result to take into account the minimum radius
                    if innerRadius > 0: #avoid divide by zero error
                        falloff = (falloff - innerRadius) / (1-innerRadius)
                    
                    #correct for times when we get a negative result (with an innerZone close to 1)
                    if falloff < 0:
                        falloff = 0
                    
                    #smooth step the linear falloff to make smooth falloff
                    A = 1;
                    B = 0;
                    
                    #trying to decide which of these smooth steps I prefer...
                    #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                    randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                    falloff = (A * randStepValue) + (B * (1 - randStepValue))
                    
                    if invertFalloffFlag !=1:
                        falloffEnvelope = 1-falloff
                    else:
                        falloffEnvelope = falloff
                    
                    resultX *= falloffEnvelope
                    resultY *= falloffEnvelope
                    resultZ *= falloffEnvelope
    
            if (inRotDArray.length() > 0):
                resultX += rotRead.x
                resultY += rotRead.y
                resultZ += rotRead.z
               
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)

        self._previousTime = om.MTime(currentTime)
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashRotate() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashRotate.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashRotate.addAttribute( mashRotate.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    mashRotate.addAttribute ( mashRotate.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashRotate.outRotationPP = tAttr.create("outRotationPP", "ourRot", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(1)
    tAttr.setUsesArrayDataBuilder(True)
    mashRotate.addAttribute( mashRotate.outRotationPP )
    
    tAttr = om.MFnTypedAttribute()
    mashRotate.inRotationPP = tAttr.create("inRotationPP", "inRot", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashRotate.addAttribute( mashRotate.inRotationPP)
    
    uAttr = om.MFnUnitAttribute()
    mashRotate.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashRotate.addAttribute(mashRotate.aTime)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.aEnvelopeX = nAttr.create("EnvelopeX", "envX", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRotate.addAttribute( mashRotate.aEnvelopeX )
    mashRotate.attributeAffects( mashRotate.aEnvelopeX, mashRotate.outRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.aEnvelopeY = nAttr.create("EnvelopeY", "envY", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRotate.addAttribute( mashRotate.aEnvelopeY )
    mashRotate.attributeAffects( mashRotate.aEnvelopeY, mashRotate.outRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.aEnvelopeZ = nAttr.create("EnvelopeZ", "envZ", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRotate.addAttribute( mashRotate.aEnvelopeZ )
    mashRotate.attributeAffects( mashRotate.aEnvelopeZ, mashRotate.outRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.targetMode = nAttr.create ( "targetMode", "rMode", om.MFnNumericData.kBoolean, 0 )
    mashRotate.addAttribute ( mashRotate.targetMode )
    mashRotate.attributeAffects( mashRotate.targetMode, mashRotate.outRotationPP)
    
    nAttr = om.MFnNumericAttribute()
    mashRotate.targetInput = nAttr.create ( "targetInput", "tin", om.MFnNumericData.k3Float, 0 )
    nAttr.setWritable(1)
    nAttr.setReadable(1)
    nAttr.setStorable(1)
    mashRotate.addAttribute ( mashRotate.targetInput )
    mashRotate.attributeAffects( mashRotate.targetInput, mashRotate.outRotationPP)
    
    mashRotate.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashRotate.addAttribute(mashRotate.falloffObject)
    mashRotate.attributeAffects(mashRotate.falloffObject, mashRotate.outRotationPP)
    
    mAttr = om.MFnMessageAttribute()
    mashRotate.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashRotate.addAttribute( mashRotate.falloffMessage )
    
    
    mashRotate.attributeAffects( mashRotate.inputArray, mashRotate.outRotationPP)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashRotateId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.deregisterNode( mashRotateId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    