# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as mpx
import maya.OpenMayaRender as mr
import maya.OpenMayaUI as omUI
import maya.mel as mel

nodeTypeName = "MASH_Points"
nodeTypeId = om.MTypeId(0x0011BE33)

glRenderer = mr.MHardwareRenderer.theRenderer()
glFT = glRenderer.glFunctionTable()

class mashPoints( mpx.MPxLocatorNode ):

    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_PointsTemplate( string $nodeName )
        {
        
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Points" "Points";
        
        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Point Node" -collapse 0;
                
        editorTemplate  -label "Enable Points" -addControl "Enable";
        //editorTemplate  -label "Enable Point IDs" -addControl "EnableText";
        editorTemplate  -label "Colour" -addControl "shapeColour";
        
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        global proc headerIcons_MASH ( string $image, string $nodeType )
        {
        string $imageName[];
        tokenize($image, ".", $imageName);
        string $nodeName[];
        tokenize($nodeType, ".", $nodeName);
        
        rowLayout -numberOfColumns 3
        -columnWidth3 162 66 162
        -columnAttach 1 "both"  0
        -columnAttach 2 "both"  0
        -columnAttach 3 "both"  0
        iconLayout_MASH;
        
        text -l $nodeName[1] nodeNameText;
        image -h 65 -w 65 -image $imageName[1] pointsImage;
        setParent..;
        
        }
        
        global proc headerIconsEdit_MASH ( string $image, string $nodeType  )
        {
        string $imageName[];
        tokenize($image, ".", $imageName);
        string $nodeName[];
        tokenize($nodeType, ".", $nodeName);
        text -e -l $nodeName[1] nodeNameText;
        image -e -h 65 -w 65 -image $imageName[1] pointsImage;
        setParent..;
        }
        ''')
    
    def __init__( self ):
        mpx.MPxLocatorNode.__init__( self )
        self.inDArray = om.MVectorArray()


    
    
    # override to set the locator name correctly
    def postConstructor(self):
        this_object		= self.thisMObject()
        fnThisNode	= om.MFnDependencyNode(this_object)
        fnThisNode.setName("MASH_PointsShape#")
    
    def compute( self, plug, data ):
        return om.kUnknownParameter

    def draw( self, view, path, style, status ):
        #get this node
        this_object = self.thisMObject()
        fnThisNode = om.MFnDependencyNode(this_object)
        locName = fnThisNode.name()
        
        #get in array
        plug = om.MPlug(this_object, self.positions)
        
        oData = plug.asMObject() 
        inDArrFn = om.MFnVectorArrayData(oData)
        self.inDArray = inDArrFn.array()
        
        numPoints = self.inDArray.length()
        
        #get enable
        plug = om.MPlug(this_object, self.aEnable)
        aEnableFlag = plug.asBool()
        '''
        #get text enable
        plug = om.MPlug(this_object, self.aEnableText)
        aEnableTextFlag = plug.asBool()
        '''
        plug = om.MPlug(this_object, self.aColour)
        
        if aEnableFlag == 1:
            
            #there is a bug in Maya 2012 and this line of code does not work, please disable the points node (uncheck Enable Points) to work around.
            #get the colour values as floats
            colour = getFloat3PlugValue(plug)
            
            # set colour
            if status == omUI.M3dView.kDormant:
                glFT.glColor3f(colour[0],colour[1],colour[2])
            textView = omUI.M3dView.active3dView()
            
            #set the point size
            glFT.glPointSize( 4 )
        
            #draw the points
            for p in range (numPoints):
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = self.inDArray[p]
                textToPrint = "."
                textPos = om.MPoint(testRead)
                view.drawText(textToPrint, textPos, omUI.M3dView.kCenter)
                
                textToPrint = (" " + str(p))
                view.drawText(textToPrint, textPos, omUI.M3dView.kLeft)
                #Bug in Maya? Points only draw at the correct location when the node is selected. Will try this in C++
                #glFT.glBegin( mr.MGL_POINTS )
                #glFT.glVertex3f( testRead.x, testRead.y, testRead.z )
                #print "point", testRead.x, testRead.y, testRead.z
                #glFT.glEnd()
                
        #I might re-enable this in the future.
        #if aEnableTextFlag == 1:
            #draw the text
            for q in range (numPoints):
                textRead = om.MVector(0.0,0.0,0.0)
                textRead = self.inDArray[q]


            # reset opengl colours
            kLeadColor = 18 # green
            kActiveColor = 15 # white
            kActiveAffectedColor = 8  # purple
            kDormantColor = 4  # blue
            kHiliteColor = 17 # pale blue
        
            if status == omUI.M3dView.kDormant:
                glFT.glColor3f(colour[0], colour[1], colour[2])
                    
            if (status == omUI.M3dView.kLead):
                view.setDrawColor( kLeadColor, omUI.M3dView.kActiveColors )
            
            elif (status == omUI.M3dView.kActive):
                view.setDrawColor( kActiveColor, omUI.M3dView.kActiveColors )
            
            elif (status == omUI.M3dView.kActiveAffected):
                view.setDrawColor( kActiveAffectedColor, omUI.M3dView.kActiveColors )
            
            elif (status == omUI.M3dView.kDormant):
                view.setDrawColor( kDormantColor, omUI.M3dView.kActiveColors )
            
            elif (status == omUI.M3dView.kHilite):
                view.setDrawColor( kHiliteColor, omUI.M3dView.kActiveColors )
        
        #finish him
        if (self.inDArray.length() > 0):
            view.endGL()

def nodeCreator():
    return mpx.asMPxPtr( mashPoints() )


def nodeInitializer():

    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    defaultIntArray = om.MIntArray();
    dIntArrayDataFn = om.MFnIntArrayData();
    dIntArrayDataFn.create( defaultIntArray );
    
    nAttr = om.MFnNumericAttribute()
    tAttr = om.MFnTypedAttribute()
    
    nAttr = om.MFnNumericAttribute()
    mashPoints.aEnable = nAttr.create ( "Enable", "ena", om.MFnNumericData.kBoolean, 1 )
    mashPoints.addAttribute ( mashPoints.aEnable )
    '''#I might re-enable this when I shift the plug-in to C++
    nAttr = om.MFnNumericAttribute()
    mashPoints.aEnableText = nAttr.create ( "EnableText", "ente", om.MFnNumericData.kBoolean, 0 )
    mashPoints.addAttribute ( mashPoints.aEnableText )
    '''
    mashPoints.positions = tAttr.create("positions", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashPoints.addAttribute( mashPoints.positions)
    
    mashPoints.ids = tAttr.create("ids", "id", om.MFnData.kIntArray, dIntArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashPoints.addAttribute( mashPoints.ids)

    nAttr = om.MFnNumericAttribute()
    mashPoints.aColour = nAttr.createColor("shapeColour", "sc")
    nAttr.setDefault(0.7, 0.85, 1.0)
    nAttr.setKeyable(1)
    nAttr.setStorable(1)
    nAttr.setUsedAsColor(1)
    nAttr.setReadable(1)
    nAttr.setWritable(1)
    mashPoints.addAttribute(mashPoints.aColour)


def initializePlugin( obj ):
    plugin = mpx.MFnPlugin( obj )
    try:
        plugin.registerNode( nodeTypeName, nodeTypeId, nodeCreator, nodeInitializer, mpx.MPxNode.kLocatorNode )
    except:
        sys.stderr.write( "Failed to register node: %s" % nodeTypeName )
        raise


def uninitializePlugin( obj ):
    plugin = mpx.MFnPlugin( obj )
    try:
        plugin.deregisterNode( nodeTypeId )
    except:
        sys.stderr.write( "Failed to deregister node: %s" % nodeTypeName )
        raise

def getFloat3PlugValue(plug):
    
    # Retrieve the value as an MObject
    object = plug.asMObject()
    
    # Convert the MObject to a float3
    numDataFn = om.MFnNumericData(object)
    xParam = om.MScriptUtil()
    xParam.createFromDouble(0.0)
    xPtr = xParam.asFloatPtr()
    yParam = om.MScriptUtil()
    yParam.createFromDouble(0.0)
    yPtr = yParam.asFloatPtr()
    zParam = om.MScriptUtil()
    zParam.createFromDouble(0.0)
    zPtr = zParam.asFloatPtr()
    numDataFn.getData3Float(xPtr, yPtr, zPtr)
    return om.MFloatVector(
                           om.MScriptUtil(xPtr).asFloat(),
                           om.MScriptUtil(yPtr).asFloat(),
                           om.MScriptUtil(zPtr).asFloat()
                           )