# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Random"

mashRandomId = om.MTypeId(0x0011BE29)

class mashRandom( omMPx.MPxNode ):
    
    inputs = om.MObject()
    outputArray = om.MObject()
    
    #AETemplate in the .py, where it should be
    mel.eval('''
        global proc AEMASH_RandomTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Random" "Random";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Random Node" -collapse 0;
        
        editorTemplate  -label "Minimum Number" -addControl "minNumber";
        editorTemplate  -label "Maximum Number" -addControl "maxNumber";
        editorTemplate  -label "Bias Extremeties" -addControl "biasHighNumbers";
        editorTemplate  -label "Random Seed" -addControl "randomSeed";
        editorTemplate -addSeparator;
        
        editorTemplate  -label "Whole Numbers" -addControl "wholeNumbers";
        editorTemplate  -label "Uniform Random" -addControl "uniformRandom";
        
        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Strength Controls" -collapse 0;
        
        editorTemplate  -label "Strength X" -addControl "EnvelopeX";
        editorTemplate  -label "Random Strength X" -addControl "randEnvelopeX";
        
        editorTemplate -addSeparator;
        
        editorTemplate  -label "Strength Y" -addControl "EnvelopeY";
        editorTemplate  -label "Random Strength Y" -addControl "randEnvelopeY";
        
        editorTemplate -addSeparator;
        
        editorTemplate  -label "Strength Z" -addControl "EnvelopeZ";
        editorTemplate  -label "Random Strength Z" -addControl "randEnvelopeZ";
        
        editorTemplate -endLayout;
        
        editorTemplate -suppress "inIterations";
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }

        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):
    
        if plug != mashRandom.outputArray:
            return om.kUnknownParameter

        outResultArray = om.MVectorArray()
        outResultArray.clear()
        
        outDoubleArray = om.MDoubleArray()
        outDoubleArray.clear()

        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashRandom.inputArray)
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashRandom.outputArray)

        #interface values
        numberOfCalcs = dataBlock.inputValue(mashRandom.inIterations).asInt()
        
        minNumber = dataBlock.inputValue(mashRandom.minNumber).asFloat()
        maxNumber = dataBlock.inputValue(mashRandom.maxNumber).asFloat()
        biasValue = dataBlock.inputValue(mashRandom.biasHighNumbers).asFloat()
        seedValue = dataBlock.inputValue(mashRandom.randomSeed).asInt()
        
        envXValue = dataBlock.inputValue(mashRandom.aEnvelopeX).asFloat()
        envYValue = dataBlock.inputValue(mashRandom.aEnvelopeY).asFloat()
        envZValue = dataBlock.inputValue(mashRandom.aEnvelopeZ).asFloat()
        randEnvXValue = dataBlock.inputValue(mashRandom.aRandEnvelopeX).asFloat()
        randEnvYValue = dataBlock.inputValue(mashRandom.aRandEnvelopeY).asFloat()
        randEnvZValue = dataBlock.inputValue(mashRandom.aRandEnvelopeZ).asFloat()
        
        randomTypeFlag = dataBlock.inputValue(mashRandom.randomType).asBool()
        uniformRandomFlag = dataBlock.inputValue(mashRandom.uniformRandom).asBool()
        
        dataType = om.MFnData.kVectorArray
        #Organise the input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #ping time, just doing this to trigger a calculation - needed when you add objects/points to the MASH Network.
        currentTime = dataBlock.inputValue(self.aTime).asTime()
        
        testRead = om.MVector(0.0,0.0,0.0)

        #Set the outputArray
        outDData = dataBlock.outputValue(mashRandom.outputArray)
        nData = om.MFnVectorArrayData()
        
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        #get the falloff object's position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
        
        if inputs.length() > 0:
            
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
    
        for i in range(numberOfCalcs):
            testRead = om.MVector(0.0,0.0,0.0)
            if (inDArray.length() > 0):
                testRead = inDArray[i]
            

            
            random.seed(i*seedValue)
            
            randStepValueX = (randEnvXValue) * (randEnvXValue) * (3 - 2 * (randEnvXValue))
            randStepValueY = (randEnvYValue) * (randEnvYValue) * (3 - 2 * (randEnvYValue))
            randStepValueZ = (randEnvZValue) * (randEnvZValue) * (3 - 2 * (randEnvZValue))
            
            #randomise the smoothstep so that each object starts and ends at a different time.
            for x in range(0,20):
                whichSmoother = random.randint(0,20)
                if (whichSmoother == 0):
                    #finishes earlier
                    randStepValueX = 1 - (1 - randStepValueX) * (1 - randStepValueX);
                if (whichSmoother == 1):
                    #starts later
                    randStepValueX = randStepValueX * randStepValueX;
            
            A = 1;
            B = 0;
            
            #finish off the smoothstep
            newRandStepValueX = (A * randStepValueX) + (B * (1 - randStepValueX))
            
            #randomise the smoothstep so that each object starts and ends at a different time.
            for x in range(0,20):
                whichSmoother = random.randint(0,20)
                if (whichSmoother == 0):
                    #finishes earlier
                    randStepValueY = 1 - (1 - randStepValueY) * (1 - randStepValueY);
                if (whichSmoother == 1):
                    #starts later
                    randStepValueY = randStepValueY * randStepValueY;

            A = 1;
            B = 0;

            #finish off the smoothstep
            newRandStepValueY = (A * randStepValueY) + (B * (1 - randStepValueY))

            #randomise the smoothstep so that each object starts and ends at a different time.
            for x in range(0,20):
                whichSmoother = random.randint(0,20)
                if (whichSmoother == 0):
                    #finishes earlier
                    randStepValueZ = 1 - (1 - randStepValueZ) * (1 - randStepValueZ);
                if (whichSmoother == 1):
                    #starts later
                    randStepValueZ = randStepValueZ * randStepValueZ;

            A = 1;
            B = 0;

            #finish off the smoothstep
            newRandStepValueZ = (A * randStepValueZ) + (B * (1 - randStepValueZ))
            
            
            
            resultX = random.uniform(minNumber,maxNumber)
            
            if (uniformRandomFlag==0):
                resultY = ((random.uniform(minNumber,maxNumber))*envYValue)*newRandStepValueY
                resultZ = ((random.uniform(minNumber,maxNumber))*envZValue)*newRandStepValueZ
                resultX = ((resultX)*envXValue)*newRandStepValueX
            else: 
                resultY = ((resultX)*envYValue)*newRandStepValueY
                resultZ = ((resultX)*envZValue)*newRandStepValueZ
                resultX = ((resultX)*envXValue)*newRandStepValueX
                
            inNormalArray = om.MFloatArray()
            inNormalArray.clear()
             
            if (randomTypeFlag!=0):
                resultX = round(resultX)
                resultY = round(resultY)
                resultZ = round(resultZ)
        
            if useFalloffFlag==1:
                if inputs.length() > 0:
                    if radius == 0:
                        radius = 0.001 #avoid divide by zero error
                    
                    #how far is our object from the sphere centre?
                    #fast approximate method, not really good enough.
                    '''
                    distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                    distanceSqrd *= 0.1
                    '''
                    #slow more accurate method
                    xd = testRead.x - falloffObjPos.x
                    yd = testRead.y - falloffObjPos.y
                    zd = testRead.z - falloffObjPos.z
                    distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                    
                    #linear falloff
                    falloff = min(1, max(0, distanceSqrd/radius))
                    
                    #normalise the result to take into account the minimum radius
                    if innerRadius > 0: #avoid divide by zero error
                        falloff = (falloff - innerRadius) / (1-innerRadius)
                    
                    #correct for times when we get a negative result (with an innerZone close to 1)
                    if falloff < 0:
                        falloff = 0
                    
                    #smooth step the linear falloff to make smooth falloff
                    A = 1;
                    B = 0;
                    
                    #trying to decide which of these smooth steps I prefer...
                    #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                    randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                    falloff = (A * randStepValue) + (B * (1 - randStepValue))
                    
                    if invertFalloffFlag !=1:
                        falloffEnvelope = 1-falloff
                    else:
                        falloffEnvelope = falloff
                    
                    resultX *= falloffEnvelope
                    resultY *= falloffEnvelope
                    resultZ *= falloffEnvelope
            
            idResult = resultX    
            outDoubleArray.append(idResult)  
                                              
            if (inDArray.length() > 0):
                resultX += testRead.x
                resultY += testRead.y
                resultZ += testRead.z
                
              
            thisResult = om.MVector(resultX,resultY,resultZ)
            outResultArray.append(thisResult)

        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)

        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashRandom() )

def nodeInitializer():
        
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    defaultDoubleArray = om.MDoubleArray();
    doubleArrayDataFn = om.MFnDoubleArrayData();
    doubleArrayDataFn.create( defaultDoubleArray );
    
    tAttr = om.MFnTypedAttribute()
    mashRandom.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashRandom.addAttribute( mashRandom.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.inIterations = nAttr.create ( "inIterations", "cal", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(0)
    mashRandom.addAttribute ( mashRandom.inIterations )
    
    tAttr = om.MFnTypedAttribute()
    mashRandom.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(0)
    tAttr.setUsesArrayDataBuilder(True)
    mashRandom.addAttribute( mashRandom.outputArray )
    
    uAttr = om.MFnUnitAttribute()
    mashRandom.aTime = uAttr.create("time", "ti", om.MFnUnitAttribute.kTime, 0.0)
    mashRandom.addAttribute(mashRandom.aTime)
    mashRandom.attributeAffects(mashRandom.aTime, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.minNumber = nAttr.create ( "minNumber", "min", om.MFnNumericData.kFloat, -10 )
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    nAttr.setHidden(0)
    nAttr.setSoftMin(-100)
    nAttr.setSoftMax(100)
    mashRandom.addAttribute ( mashRandom.minNumber )
    mashRandom.attributeAffects( mashRandom.minNumber, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.maxNumber = nAttr.create ( "maxNumber", "max", om.MFnNumericData.kFloat, 10 )
    nAttr.setStorable(1)
    nAttr.setSoftMin(-100)
    nAttr.setSoftMax(100)
    nAttr.setKeyable(1)
    nAttr.setHidden(0)
    mashRandom.addAttribute ( mashRandom.maxNumber )
    mashRandom.attributeAffects( mashRandom.maxNumber, mashRandom.outputArray)

    
    nAttr = om.MFnNumericAttribute()
    mashRandom.biasHighNumbers = nAttr.create ( "biasHighNumbers", "biasH", om.MFnNumericData.kFloat, 10 )
    nAttr.setStorable(1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    nAttr.setHidden(0)
    mashRandom.addAttribute ( mashRandom.biasHighNumbers )
    mashRandom.attributeAffects( mashRandom.biasHighNumbers, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.randomSeed = nAttr.create ( "randomSeed", "seed", om.MFnNumericData.kInt, 1 )
    nAttr.setStorable(1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setKeyable(1)
    tAttr.setHidden(0)
    mashRandom.addAttribute ( mashRandom.randomSeed )
    mashRandom.attributeAffects( mashRandom.randomSeed, mashRandom.outputArray)

    nAttr = om.MFnNumericAttribute()
    mashRandom.randomType = nAttr.create ( "wholeNumbers", "whoNum", om.MFnNumericData.kBoolean, 0 )
    mashRandom.addAttribute ( mashRandom.randomType )
    mashRandom.attributeAffects( mashRandom.randomType, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.uniformRandom = nAttr.create ( "uniformRandom", "uniRand", om.MFnNumericData.kBoolean, 0 )
    mashRandom.addAttribute ( mashRandom.uniformRandom )
    mashRandom.attributeAffects( mashRandom.uniformRandom, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.aEnvelopeX = nAttr.create("EnvelopeX", "envX", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aEnvelopeX )
    mashRandom.attributeAffects( mashRandom.aEnvelopeX, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.aRandEnvelopeX = nAttr.create("randEnvelopeX", "raEnX", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aRandEnvelopeX )
    mashRandom.attributeAffects( mashRandom.aRandEnvelopeX, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.aEnvelopeY = nAttr.create("EnvelopeY", "envY", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aEnvelopeY )
    mashRandom.attributeAffects( mashRandom.aEnvelopeY, mashRandom.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashRandom.aRandEnvelopeY = nAttr.create("randEnvelopeY", "raEnY", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aRandEnvelopeY )
    mashRandom.attributeAffects( mashRandom.aRandEnvelopeY, mashRandom.outputArray)

    nAttr = om.MFnNumericAttribute()
    mashRandom.aEnvelopeZ = nAttr.create("EnvelopeZ", "envZ", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aEnvelopeZ )
    mashRandom.attributeAffects( mashRandom.aEnvelopeZ, mashRandom.outputArray)

    nAttr = om.MFnNumericAttribute()
    mashRandom.aRandEnvelopeZ = nAttr.create("randEnvelopeZ", "raEnZ", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashRandom.addAttribute( mashRandom.aRandEnvelopeZ )
    mashRandom.attributeAffects( mashRandom.aRandEnvelopeZ, mashRandom.outputArray)
    

    mashRandom.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashRandom.addAttribute(mashRandom.falloffObject)
    mashRandom.attributeAffects(mashRandom.falloffObject, mashRandom.outputArray)
    
    mAttr = om.MFnMessageAttribute()
    mashRandom.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashRandom.addAttribute( mashRandom.falloffMessage )


    mashRandom.attributeAffects( mashRandom.inputArray, mashRandom.outputArray)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashRandomId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashRandomId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    