# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/
import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Trig"

mashTrigId = om.MTypeId(0x0011BE87)

class mashTrig( omMPx.MPxNode ):
    
    time = om.MObject()
    outputArray   = om.MObject()
    inputArray   = om.MObject()
    
    #AETemplate with radio buttons, swish.
    mel.eval('''
        global proc AEMASH_TrigTemplate( string $nodeName )
            {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Trig" "Trig";

            editorTemplate -beginScrollLayout;
            editorTemplate -beginLayout "Trig Node" -collapse 0;
        
            editorTemplate  -label "Amplitude X" -addControl "amplitudeX";
            editorTemplate  -label "Amplitude Y" -addControl "amplitudeY";
            editorTemplate  -label "Amplitude Z" -addControl "amplitudeZ";
            editorTemplate -addSeparator;
            editorTemplate  -label "Frequency X" -addControl "frequencyX";
            editorTemplate  -label "Frequency Y" -addControl "frequencyY";
            editorTemplate  -label "Frequency Z" -addControl "frequencyZ";
            
            editorTemplate -addSeparator;
            editorTemplate  -label "Strength" -addControl "Envelope";
            editorTemplate  -label "Random Strength" -addControl "randEnvelope";
            editorTemplate  -label "Step Strength" -addControl "StepEnvelope";

            editorTemplate -suppress "inIterations";
            editorTemplate -addSeparator;
            editorTemplate -callCustom "xAxisNew"
            "xAxisReplace"
            "xAxis";
            
            editorTemplate -callCustom "yAxisNew"
            "yAxisReplace"
            "yAxis";
            
            editorTemplate -callCustom "zAxisNew"
            "zAxisReplace"
            "zAxis";

            
            editorTemplate -addSeparator;
            
            editorTemplate  -label "Step" -addControl "step";
            editorTemplate  -label "Pattern" -addControl "modular";
            editorTemplate  -label "Absolute Values" -addControl "positiveNoise";
            
            editorTemplate -beginLayout "Falloff Object" -collapse 1;
            //the custom command this creates is in the Waiter node's AE Template.
            editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
            editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
            editorTemplate -endLayout;
        
            editorTemplate -beginLayout "Inputs" -collapse 1;
            editorTemplate  -label "Time" -addControl "time";
            editorTemplate  -label "Offsets" -addControl "offset";
            editorTemplate -endLayout;
            
            
            AEdependNodeTemplate $nodeName;
        
            editorTemplate -addExtraControls;
            editorTemplate -endScrollLayout;
        }
        
        global proc xAxisNew( string $attrName )
        {
        // Maya the "quick set" control for the funciton attribute
        
        radioButtonGrp
		-label "X Function: "
		-numberOfRadioButtons 3
		-label1 "Sin"
		-data1 1
		-label2 "Cos"
		-data2 2
		-label3 "Tan"
        -data3 3
		xAxisGrp;
        connectControl xAxisGrp $attrName;

        }
        
        global proc xAxisReplace( string $attrName )
        {

        connectControl xAxisGrp $attrName;
    
        }
        
        global proc yAxisNew( string $attrName )
        {
        // Maya the "quick set" control for the funciton attribute
        radioButtonGrp
		-label "Y Function: "
		-numberOfRadioButtons 3
		-label1 "Sin"
		-data1 1
		-label2 "Cos"
		-data2 2
		-label3 "Tan"
        -data3 3
		yAxisGrp;
        connectControl yAxisGrp $attrName;
        
        }
        
        global proc yAxisReplace( string $attrName )
        {

        connectControl yAxisGrp $attrName;
        
        }
        
        global proc zAxisNew( string $attrName )
        {
        // Maya the "quick set" control for the funciton attribute
        radioButtonGrp
		-label "Z Function: "
		-numberOfRadioButtons 3
		-label1 "Sin"
		-data1 1
		-label2 "Cos"
		-data2 2
		-label3 "Tan"
        -data3 3
		zAxisGrp;
        connectControl zAxisGrp $attrName;
        
        }
        
        global proc zAxisReplace( string $attrName )
        {
        // Install the connection between the radioButtonGrp and the
        // actual scale attribute
        connectControl zAxisGrp $attrName;
        
        }

        
        ''')
    
    def __init__( self ):
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):
        if plug != mashTrig.outputArray:
            return om.kUnknownParameter    
            
        # Get the time
        dataHandle = dataBlock.inputValue(mashTrig.time).asFloat()
        
        #Get input array
        inputArray_dataHandle = dataBlock.inputValue(mashTrig.inputArray)
        
        #Get output array
        outputArray_dataHandle = dataBlock.outputValue(mashTrig.outputArray)
        
        #get the interface settings
        amplitudeValueX = dataBlock.inputValue(mashTrig.aAmplitudeX).asFloat()
        frequencyValueX = dataBlock.inputValue(mashTrig.aFrequencyX).asFloat()
        amplitudeValueY = dataBlock.inputValue(mashTrig.aAmplitudeY).asFloat()
        frequencyValueY = dataBlock.inputValue(mashTrig.aFrequencyY).asFloat()
        amplitudeValueZ = dataBlock.inputValue(mashTrig.aAmplitudeZ).asFloat()
        frequencyValueZ = dataBlock.inputValue(mashTrig.aFrequencyZ).asFloat()
        stepValue = dataBlock.inputValue(mashTrig.aStep).asFloat()
        funcXflag = dataBlock.inputValue(mashTrig.aXaxis).asInt()
        funcYflag = dataBlock.inputValue(mashTrig.aYaxis).asInt()
        funcZflag = dataBlock.inputValue(mashTrig.aZaxis).asInt()
        modularValue = dataBlock.inputValue(mashTrig.aModular).asInt()
        numberOfCalcs = dataBlock.inputValue(mashTrig.inIterations).asInt()
        positiveFlag = dataBlock.inputValue(self.aPositive).asBool()
        EnvelopeValue = dataBlock.inputValue(mashTrig.aEnvelope).asFloat()
        EnvelopeRandValue = dataBlock.inputValue(mashTrig.aRandEnvelope).asFloat()
        stepEnvelopeValue = dataBlock.inputValue(self.aStepEnvelope).asFloat()
        offsetValueArray = dataBlock.inputValue(mashTrig.aOffset).asFloat3()
        
        #create output array
        outResultArray = om.MVectorArray()
        outResultArray.clear()

        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
       
        #Set the outputArray
        outDData = dataBlock.outputValue(self.outputArray)
        nData = om.MFnVectorArrayData()       
        
        #get the falloff object position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
            
        #check for falloff object
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        if inputs.length() > 0:
            
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
        
        if (dataHandle == 0):
            dataHandle = 1
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
        
        #run the calculation the required number of times
        for i in range(numberOfCalcs):
            if (inDArray.length() > 0):
                #if there's an in array, get the appropriate input value
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
            
            #set the step value
            stepAddition = stepValue*i
            
            #do the sin cos tan magic
            if (i%modularValue == 0):
                if (funcXflag == 0):
                    resultX = math.sin((dataHandle+stepAddition) * (frequencyValueX*0.1)) * amplitudeValueX
                elif (funcXflag == 1):
                    resultX = math.cos((dataHandle+stepAddition) * (frequencyValueX*0.1)) * amplitudeValueX
                else:
                    resultX = math.tan((dataHandle+stepAddition) * (frequencyValueX*0.1)) * amplitudeValueX
                
                if (funcYflag == 0):
                    resultY = math.sin((dataHandle+stepAddition) * (frequencyValueY*0.1)) * amplitudeValueY
                elif (funcYflag == 1):
                    resultY = math.cos((dataHandle+stepAddition) * (frequencyValueY*0.1)) * amplitudeValueY
                else:        
                    resultY = math.tan((dataHandle+stepAddition) * (frequencyValueY*0.1)) * amplitudeValueY
                        
                if (funcZflag == 0):
                    resultZ = math.sin((dataHandle+stepAddition) * (frequencyValueZ*0.1)) * amplitudeValueZ
                elif (funcZflag == 1):
                    resultZ = math.cos((dataHandle+stepAddition) * (frequencyValueZ*0.1)) * amplitudeValueZ
                else:
                    resultZ = math.tan((dataHandle+stepAddition) * (frequencyValueZ*0.1)) * amplitudeValueZ
                
                
                if (positiveFlag!=0):
                    if (resultX < 0):
                        resultX = math.fabs(resultX)
                    if (resultY < 0):
                        resultY = math.fabs(resultY)
                    if (resultZ < 0):
                        resultZ = math.fabs(resultZ)
                
                if (EnvelopeRandValue < 1):
                    randStepValue = (EnvelopeRandValue) * (EnvelopeRandValue) * (3 - 2 * (EnvelopeRandValue))
                    
                    #randomise the smoothstep so that each object starts and ends at a different time.
                    for x in range(0,20):
                        whichSmoother = random.randint(0,20)
                        if (whichSmoother == 0):
                            #finishes earlier
                            randStepValue = 1 - (1 - randStepValue) * (1 - randStepValue);
                        if (whichSmoother == 1):
                            #starts later
                            randStepValue = randStepValue * randStepValue;
                    
                    A = 1;
                    B = 0;
                    
                    #finish off the smoothstep
                    newRandStepValue = (A * randStepValue) + (B * (1 - randStepValue))
                else:
                    newRandStepValue = EnvelopeRandValue
                
                stepResult = 1
                if (stepEnvelopeValue < 1):
                    #find the scale range for this i
                    step = (1.0/(numberOfCalcs+1))*(i+1)
                    min = (1.0/(numberOfCalcs+1))*i
                    
                    #are you step enviloping?
                    if stepEnvelopeValue <= step:
                        #normalise to get the correct scale value
                        stepResult = (stepEnvelopeValue-min)/(step-min)
                    else:
                        stepResult = 1.0
                    
                    if stepResult < step:
                        stepResult = 0.0

                #factor in Envelopes and offsets
                resultX = ((resultX+offsetValueArray[0])*EnvelopeValue)*newRandStepValue*stepResult
                resultY = ((resultY+offsetValueArray[1])*EnvelopeValue)*newRandStepValue*stepResult
                resultZ = ((resultZ+offsetValueArray[2])*EnvelopeValue)*newRandStepValue*stepResult
                
                if useFalloffFlag==1:
                    if inputs.length() > 0:
                        if radius == 0:
                            radius = 0.001 #avoid divide by zero error
                        
                        #how far is our object from the sphere centre?
                        #fast approximate method, not really good enough.
                        '''
                            distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                            distanceSqrd *= 0.1
                            '''
                        #slow more accurate method
                        xd = testRead.x - falloffObjPos.x
                        yd = testRead.y - falloffObjPos.y
                        zd = testRead.z - falloffObjPos.z
                        distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                        
                        #linear falloff
                        falloff = min(1, max(0, distanceSqrd/radius))
                        
                        #normalise the result to take into account the minimum radius
                        if innerRadius > 0: #avoid divide by zero error
                            falloff = (falloff - innerRadius) / (1-innerRadius)
                        
                        #correct for times when we get a negative result (with an innerZone close to 1)
                        if falloff < 0:
                            falloff = 0
                        
                        #smooth step the linear falloff to make smooth falloff
                        A = 1;
                        B = 0;
                        
                        #trying to decide which of these smooth steps I prefer...
                        #randStepValue = falloff * falloff * falloff * (falloff * (falloff * 6.0 - 15.0) + 10.0); #smoother, perhaps too smooth...
                        randStepValue = (falloff) * (falloff) * (3 - 2 * (falloff))
                        falloff = (A * randStepValue) + (B * (1 - randStepValue))
                        
                        if invertFalloffFlag !=1:
                            falloffEnvelope = 1-falloff
                        else:
                            falloffEnvelope = falloff
                        
                        resultX *= falloffEnvelope
                        resultY *= falloffEnvelope
                        resultZ *= falloffEnvelope
                
            #if there's an input array, add the values we got above
            if (inDArray.length() > 0):
                resultX +=  testRead.x 
                resultY +=  testRead.y 
                resultZ +=  testRead.z 
                
            
            
            #get the result as a vector
            thisResult = om.MVector(resultX,resultY,resultZ)
            
            #add it to the output array
            outResultArray.append(thisResult)
        
        #set the output array
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        
        #clean the plughole
        dataBlock.setClean(plug)
        

def nodeCreator():
    
    return omMPx.asMPxPtr( mashTrig() )

def nodeInitializer():
    
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    tAttr = om.MFnTypedAttribute()
    mashTrig.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setReadable(True)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashTrig.addAttribute( mashTrig.inputArray)
    
    
    tAttr = om.MFnTypedAttribute()
    mashTrig.outputArray = tAttr.create("outputArray", "outArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(True)
    tAttr.setKeyable(False)
    tAttr.setReadable(True)
    tAttr.setHidden(0)
    tAttr.setUsesArrayDataBuilder(True)
    mashTrig.addAttribute( mashTrig.outputArray )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(0)
    mashTrig.addAttribute ( mashTrig.inIterations )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.time = nAttr.create ( "time", "tm", om.MFnNumericData.kFloat, 1 )
    nAttr.setStorable(1)
    nAttr.setKeyable(1)
    mashTrig.addAttribute ( mashTrig.time )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aOffset = nAttr.create( "offset", "off", om.MFnNumericData.k3Float, 0.0 )
    nAttr.setStorable(1)
    nAttr.setWritable(1)
    mashTrig.addAttribute( mashTrig.aOffset )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aStep = nAttr.create("step", "st", om.MFnNumericData.kFloat, 1.0)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aStep )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aModular = nAttr.create("modular", "mod", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(20)
    mashTrig.addAttribute( mashTrig.aModular )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aAmplitudeX = nAttr.create("amplitudeX", "ampX", om.MFnNumericData.kFloat, 5)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aAmplitudeX )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aFrequencyX = nAttr.create("frequencyX", "freqX", om.MFnNumericData.kFloat, 0.1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aFrequencyX )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aAmplitudeY = nAttr.create("amplitudeY", "ampY", om.MFnNumericData.kFloat, 5)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aAmplitudeY )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aFrequencyY = nAttr.create("frequencyY", "freqY", om.MFnNumericData.kFloat, 0.1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aFrequencyY )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aAmplitudeZ = nAttr.create("amplitudeZ", "ampZ", om.MFnNumericData.kFloat, 5)
    nAttr.setSoftMin(-20)
    nAttr.setSoftMax(20)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aAmplitudeZ )
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aFrequencyZ = nAttr.create("frequencyZ", "freqZ", om.MFnNumericData.kFloat, 0.1)
    nAttr.setSoftMin(0)
    nAttr.setSoftMax(2)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aFrequencyZ )

    nAttr = om.MFnNumericAttribute()
    mashTrig.aEnvelope = nAttr.create("Envelope", "env", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aEnvelope )
    mashTrig.attributeAffects( mashTrig.aEnvelope, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aRandEnvelope = nAttr.create("randEnvelope", "raEn", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aRandEnvelope )
    mashTrig.attributeAffects( mashTrig.aRandEnvelope, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aStepEnvelope = nAttr.create("StepEnvelope", "StEnv", om.MFnNumericData.kFloat, 1)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashTrig.addAttribute( mashTrig.aStepEnvelope )
    mashTrig.attributeAffects( mashTrig.aStepEnvelope, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aXaxis = nAttr.create("xAxis", "xAx", om.MFnNumericData.kInt, 1)
    mashTrig.addAttribute( mashTrig.aXaxis )
    mashTrig.attributeAffects( mashTrig.aXaxis, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aYaxis = nAttr.create("yAxis", "yAx", om.MFnNumericData.kInt, 1)
    mashTrig.addAttribute( mashTrig.aYaxis )
    mashTrig.attributeAffects( mashTrig.aYaxis, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aZaxis = nAttr.create("zAxis", "zAx", om.MFnNumericData.kInt, 1)
    mashTrig.addAttribute( mashTrig.aZaxis )
    mashTrig.attributeAffects( mashTrig.aZaxis, mashTrig.outputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashTrig.aPositive = nAttr.create("positiveNoise", "posNoi", om.MFnNumericData.kBoolean,0)
    mashTrig.addAttribute( mashTrig.aPositive )
    mashTrig.attributeAffects( mashTrig.aPositive, mashTrig.outputArray)
    
    mashTrig.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashTrig.addAttribute(mashTrig.falloffObject)
    mashTrig.attributeAffects(mashTrig.falloffObject, mashTrig.outputArray)
    
    mAttr = om.MFnMessageAttribute()
    mashTrig.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashTrig.addAttribute( mashTrig.falloffMessage )
    
    mashTrig.attributeAffects( mashTrig.aModular, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.time, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aStep, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aOffset, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aAmplitudeX, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aFrequencyX, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aAmplitudeY, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aFrequencyY, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aAmplitudeZ, mashTrig.outputArray)
    mashTrig.attributeAffects( mashTrig.aFrequencyZ, mashTrig.outputArray)
    

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashTrigId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashTrigId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    