# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel

kPluginNodeTypeName = "MASH_Visibility"

mashVisibilityId = om.MTypeId(0x87012)

class mashVisibility( omMPx.MPxNode ):
    
    inputs = om.MObject()
    outVisiblityPP = om.MObject()
    
    #AETemplate in the .py
    mel.eval('''
        global proc AEMASH_VisibilityTemplate( string $nodeName )
        {
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Visibility" "Visibility";

        editorTemplate -beginScrollLayout;
		editorTemplate -beginLayout "Visibility Node" -collapse 0;
        
        editorTemplate  -label "Strength" -addControl "strength";
        editorTemplate -addSeparator;
        editorTemplate  -label "Randomise" -addControl "randomStrength";
        editorTemplate  -label "Random Seed" -addControl "seed";
        
        editorTemplate -beginLayout "Falloff Object" -collapse 1;
        //the custom command this creates is in the Waiter node's AE Template.
        editorTemplate -callCustom "createInputs_MASH" "editInputs_MASH" "falloffObject" "Falloff Object";
        editorTemplate -callCustom "AEfalloffButton" "AEfalloffButtonEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -suppress "inIterations";
        
        
        editorTemplate -endLayout;
		editorTemplate -endLayout;
        
		AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
        
        
        ''')
        
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )

    def compute(self, plug, dataBlock):

        #only calculate if the out visibility is hooked up
        if plug != mashVisibility.outVisibilityPP:
            return om.kUnknownParameter
        
        outResultArray = om.MDoubleArray()
        outResultArray.clear()

        #how many times to run the code
        numberOfCalcs = dataBlock.inputValue(mashVisibility.inIterations).asInt()
        
        #get values from the interface
        seedValue = dataBlock.inputValue(mashVisibility.aSeed).asInt()
        strengthValue = dataBlock.inputValue(mashVisibility.aStrength).asFloat()
        randomFlag = dataBlock.inputValue(mashVisibility.aRandomStrength).asBool()

        #Organise the input array
        inDArray = om.MVectorArray()
        inDArray.clear()
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #set up a data object for the array data 
        outDData = dataBlock.outputValue(mashVisibility.outVisibilityPP)
        nData = om.MFnDoubleArrayData()
        
        #avoid divide by zero errors while we set up
        if numberOfCalcs == 0:
                numberOfCalcs +=1
                
        calcStep = 1.0/numberOfCalcs
        
        #create empty list and fill with crap to stop maya erroring
        shuffleList = [0,0,0,0]
        del shuffleList[:]
        
        #random seed, duh
        random.seed(seedValue)
    
        #get the falloff object position
        falloffObjPos = om.MPoint(dataBlock.inputValue(self.falloffObject).asFloatVector())
        
        #check for falloff object message connection
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.falloffMessage)
        inputs = om.MPlugArray()
        inputs.clear()
        plugMessage.connectedTo( inputs, False, True )
        
        #useFalloff defaults to off
        useFalloffFlag = 0
        
        if inputs.length() > 0:
            
            #get the falloff node
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            
            setName = setDepNode.name()
            
            #get the attributes from the falloff node
            radiusAttr = setDepNode.attribute("radius")
            innerAttr = setDepNode.attribute("innerRadius")
            enableAttr = setDepNode.attribute("useFalloff")
            invertAttr = setDepNode.attribute("invertFalloff")
            
            newPlug = om.MPlug(sourceNode, radiusAttr)
            radius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, innerAttr)
            innerRadius = newPlug.asFloat()
            
            newPlug = om.MPlug(sourceNode, enableAttr)
            useFalloffFlag = newPlug.asBool()
            
            newPlug = om.MPlug(sourceNode, invertAttr)
            invertFalloffFlag = newPlug.asBool()
    
        
        #are there more items in the input array then calculations from the waiter? - This almost certainly means echo mode, so compensate for that.
        if inDArray.length() > numberOfCalcs:
            numberOfCalcs = inDArray.length()
    
        for i in range(numberOfCalcs):
            testRead = om.MVector(0.0,0.0,0.0)
            #if theres an in array, get the value
            if (inDArray.length() > 0):
                testRead = inDArray[i]
            
            #calculate visibility
            if ((calcStep*i) <= strengthValue):
                if strengthValue > 0:
                    result = 0
                else:
                    result = 1
            else:
                result = 1
            
            #if there is a falloff object...
            if useFalloffFlag==1:
                if inputs.length() > 0:
                    if radius == 0:
                        radius = 0.001 #avoid divide by zero error
                    
                    #how far is our object from the sphere centre?
                    #fast approximate method, not really good enough.
                    '''
                        distanceSqrd = (testRead.x - falloffObjPos.x)**2 + (testRead.y - falloffObjPos.y)**2 + (testRead.z - falloffObjPos.z)**2
                        distanceSqrd *= 0.1
                        '''
                    #slow more accurate method
                    xd = testRead.x - falloffObjPos.x
                    yd = testRead.y - falloffObjPos.y
                    zd = testRead.z - falloffObjPos.z
                    distanceSqrd = math.sqrt(xd*xd + yd*yd + zd*zd)
                    
                    #linear falloff
                    falloff = min(1, max(0, distanceSqrd/radius))
                    
                    if innerRadius == 1:
                        innerRadius = 0.999 #avoid divide by zero error
                    #normalise the result to take into account the minimum radius
                    if innerRadius > 0: #avoid divide by zero error
                        falloff = (falloff - innerRadius) / (1-innerRadius)
                    
                    #correct for times when we get a negative result (with an innerZone close to 1)
                    if falloff < 0:
                        falloff = 0
                    
                    if invertFalloffFlag !=1:
                        falloffEnvelope = 1-falloff
                        if falloffEnvelope < 1:
                            falloffEnvelope = 0
                        else:
                            falloffEnvelope = 1
                    else:
                        falloffEnvelope = falloff
                        if falloffEnvelope > 0:
                            falloffEnvelope = 1
                        else:
                            falloffEnvelope = 0
                    
                    result = falloffEnvelope
                    
            #add result to the list
            shuffleList.append(result)
            
        if randomFlag == 1:
            #if random is turned on, and there's no falloff object, shuffle the list
            if useFalloffFlag==0:
                random.shuffle(shuffleList)
    
        #add the list to the out array
        for i in range(numberOfCalcs):
            outResultArray.append(shuffleList[i])
    
        #set the out array
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
        
        dataBlock.setClean(plug)

        
def nodeCreator():
    
    return omMPx.asMPxPtr( mashVisibility() )

def nodeInitializer():
    
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray);
    
    defaultDoubleArray = om.MDoubleArray();
    doubleArrayDataFn = om.MFnDoubleArrayData();
    doubleArrayDataFn.create( defaultDoubleArray );
    
    tAttr = om.MFnTypedAttribute()
    mashVisibility.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(1)
    mashVisibility.addAttribute( mashVisibility.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashVisibility.inIterations = nAttr.create ( "inIterations", "inIter", om.MFnNumericData.kInt, 0 )
    nAttr.setStorable(0)
    nAttr.setKeyable(0)
    nAttr.setHidden(1)
    mashVisibility.addAttribute ( mashVisibility.inIterations )

    nAttr = om.MFnNumericAttribute()
    mashVisibility.aStrength = nAttr.create("strength", "str", om.MFnNumericData.kFloat, 0)
    nAttr.setMin(0)
    nAttr.setMax(1)
    nAttr.setKeyable(1)
    mashVisibility.addAttribute( mashVisibility.aStrength )
    
    nAttr = om.MFnNumericAttribute()
    mashVisibility.aSeed = nAttr.create("seed", "see", om.MFnNumericData.kInt, 1)
    nAttr.setMin(1)
    nAttr.setSoftMax(100)
    nAttr.setKeyable(1)
    mashVisibility.addAttribute( mashVisibility.aSeed )
    
    nAttr = om.MFnNumericAttribute()
    mashVisibility.aRandomStrength = nAttr.create ( "randomStrength", "raStr", om.MFnNumericData.kBoolean, 0 )
    mashVisibility.addAttribute ( mashVisibility.aRandomStrength )
    
    mashVisibility.falloffObject = nAttr.createPoint("falloffObject", "fallObj")
    mashVisibility.addAttribute(mashVisibility.falloffObject)
    
    
    mAttr = om.MFnMessageAttribute()
    mashVisibility.falloffMessage = mAttr.create( "falloffMessage", "fmsg" )
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashVisibility.addAttribute( mashVisibility.falloffMessage )
    
    tAttr = om.MFnTypedAttribute()
    mashVisibility.outVisibilityPP = tAttr.create("outVisibilityPP", "outV", om.MFnData.kDoubleArray, doubleArrayDataFn.object())
    mashVisibility.addAttribute( mashVisibility.outVisibilityPP )
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    
    mashVisibility.attributeAffects(mashVisibility.falloffObject, mashVisibility.outVisibilityPP)
    mashVisibility.attributeAffects( mashVisibility.aSeed, mashVisibility.outVisibilityPP)
    mashVisibility.attributeAffects( mashVisibility.aRandomStrength, mashVisibility.outVisibilityPP)
    mashVisibility.attributeAffects( mashVisibility.aStrength, mashVisibility.outVisibilityPP)
    mashVisibility.attributeAffects( mashVisibility.inputArray, mashVisibility.outVisibilityPP)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        mplugin.registerNode( kPluginNodeTypeName, mashVisibilityId, nodeCreator, nodeInitializer )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashVisibilityId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
    