# MASH Nodes by Ian Waters: http://ianwaters.posterous.com/

import sys
import math
import random
import maya.OpenMaya as om
import maya.OpenMayaMPx as omMPx
import maya.mel as mel
import maya.cmds as mc

#The purpose of this node is to dish out the values the other arrays cooked up.

kPluginNodeTypeName = "MASH_Waiter"
mashWaiterId = om.MTypeId(0x0011BE0F)


class mashWaiter( omMPx.MPxNode ):

    inputArray = om.MObject()
    output   = om.MObject()
    numberOfOutputs   = om.MObject()
    
    #AETemplate in the .py, ahhh yeahhh.
    mel.eval('''
        global proc AEMASH_WaiterTemplate( string $nodeName )
        {
        //this callcustom is stored in the points node.
        editorTemplate -callCustom "headerIcons_MASH" "headerIconsEdit_MASH" "ae_MASH_Waiter" "Waiter";

        editorTemplate -beginScrollLayout;
        editorTemplate -beginLayout "Waiter Node" -collapse 0;
              
        editorTemplate  -label "Add node to:" -addControl "chainType";
        
        editorTemplate -callCustom "AEwaiterButtons" "AEwaiterButtonsEdit" "";
        editorTemplate -callCustom "AEwaiterDummyConnections" "AEwaiterDummyConnectionsEdit" "";
        editorTemplate  -label "Number of Objects: " -addControl "numberOfOutputs";
        
        editorTemplate -beginLayout "Initial State" -collapse 0;
        editorTemplate -callCustom "AEsetInitialButtons" "AEsetInitialButtonsEdit" "";
        editorTemplate -endLayout;
        
        editorTemplate -beginLayout "Instancer Mode" -collapse 1;
        editorTemplate  -label "Instancer Mode" -addControl "outputPPMode";
        editorTemplate -callCustom "AEsetPPModeButtons" "AEsetPPModeButtonsEdit" "";
        
        editorTemplate -endLayout;

        
        AEdependNodeTemplate $nodeName;
        
        editorTemplate -addExtraControls;
        editorTemplate -endScrollLayout;
        }
 
        global proc AEsetInitialButtons ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        separator -w 100 -h 5 -hr 1 -st "none";
        
        button -label "Set Initial State"  -c ("waiterButtonCMDS " + $nodeName[0] + " 1")
        setInitialMASHButton;
        
        button -label "Clear Initial State"  -c ("waiterButtonCMDS " + $nodeName[0] + " 2")
        clearInitialMASHButton;
        
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        global proc AEsetInitialButtonsEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("waiterButtonCMDS " + $nodeName[0] + " 1")
        setInitialMASHButton;
        
        button -e -c ("waiterButtonCMDS " + $nodeName[0] + " 2")
        clearInitialMASHButton;
        }
        
        global proc AEsetPPModeButtons ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -label "Set-Up Instancer Connections"  -c ("waiterButtonCMDS " + $nodeName[0] + " 4")
        setPPModeMASHButton;
        
        rowColumnLayout -nc 1 -cw 1 390 -cal 1 "center" waiterLayout_MASH;
        
        /*//disabled but may be enabled in the future.
        checkBox -label "Position" -v 1 boxPos;
        checkBox -label "Rotation" boxRot;
        checkBox -label "Scale" boxSca;
        checkBox -label "ObjectID" boxOID;
        */
        checkBoxGrp -ncb 1
        -l "Include Object ID and Visibility"
        -v1 false
        -cw 1 275 boxVisId;
                setParent ..;
        //separator -w 390 -h 10 -hr 1 -st "single";
        separator -w 100 -h 5 -hr 1 -st "none";
        }
        
        
        global proc AEsetPPModeButtonsEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("waiterButtonCMDS " + $nodeName[0] + " 4")
        setPPModeMASHButton;
        
        }
        
        global proc AEwaiterButtons ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        setUITemplate -pst attributeEditorPresetsTemplate;
        
        rowColumnLayout -nc 1 -cw 1 390 boundingObjectsLayout_SOUP;
        
        separator -w 100 -h 5 -hr 1 -st "none";
        
        
        separator -w 100 -h 5 -hr 1 -st "none";
        rowLayout
        -numberOfColumns 2
        -columnWidth2 190 190
        -columnAlign2 "center" "center";
        
        optionMenu -label "Node Type:" nodeTypeMenu;
        menuItem -label "Curve"; 
        menuItem -label "Delay";     
        menuItem -label "Distribute"; 
        menuItem -label "ID";       
        menuItem -label "Inheritance";
        menuItem -label "Mute";
        menuItem -label "Noise";
        menuItem -label "Orient";
        menuItem -label "Random";
        menuItem -label "Spring";
        menuItem -label "Trigonometry";
        menuItem -label "Visibility";
        
        button -label "Add Node"  -c ("waiterButtonCMDS " + $nodeName[0] + " 3")
        addMASHNodeButton;
        
        setParent..;
        
        separator -w 390 -h 10 -hr 1 -st "in";
        separator -w 100 -h 5 -hr 1 -st "none";
        
        setParent ..;
        
        setUITemplate -ppt;
        }
        
        global proc AEwaiterButtonsEdit ( string $attr )
        {
        string $nodeName[];
        tokenize($attr, ".", $nodeName);
        
        button -e -c ("waiterButtonCMDS " + $nodeName[0] + " 3")
        addMASHNodeButton;

        }
        
        global proc waiterButtonCMDS (string $nodeName, int $whichCMD)
        {
        if ($whichCMD == 1)
        {
        string $selected[] = `sets -q $nodeName`;
        vector $vArray[];
        
        for ($i = 0; $i < size($selected); $i++) {
        $vArray[$i] = `getAttr ($selected[$i] + ".translate")`;
        }

        string $attr = "initialState";
        string $cmd = "setAttr -type vectorArray ";
        $cmd += ($nodeName + "." + $attr);
        $cmd += ( " " + size($vArray) + " ");
        
        for ($item in $vArray)
        $cmd += ($item + " ");
        eval($cmd);
        print "MASH_Waiter: Initial state set.";
        }
        
        if ($whichCMD == 2)
        {
        string $selected[] = `sets -q $nodeName`;
        vector $vArray[];
        
        for ($i = 0; $i < size($selected); $i++) {
        $vArray[$i] = <<0,0,0>>;
        }

        string $attr = "initialState";
        string $cmd = "setAttr -type vectorArray ";
        $cmd += ($nodeName + "." + $attr);
        $cmd += ( " " + size($vArray) + " ");
        
        for ($item in $vArray)
        $cmd += ($item + " ");
        eval($cmd);
        print "MASH_Waiter: Initial state cleared.";
        }
        
        if ($whichCMD == 4)
            {

        
            string $incoming[] = `listConnections -d 0 -s 1 ($nodeName+".positions")`;
            string $inRotation[] = `listConnections -d 0 -s 1 ($nodeName+".inRotationPP")`;
            int $loaded = `pluginInfo -q -l SOuP`;
            
            if ($loaded == 1) 
                {
                string $instancer;
                string $arrayNodePos = `createNode arrayToDynArrays -n atdaPos`;
                string $arrayNodeRot = `createNode arrayToDynArrays -n atdaRot`;
                string $arrayNodeSca = `createNode arrayToDynArrays -n atdaScale`;
                string $arrayNodeVis;
                string $arrayNodeId;
                if (`checkBoxGrp -q -v1 boxVisId`)
                {
                $arrayNodeVis = `createNode arrayToDynArrays -n atdaVis`;
                $arrayNodeId = `createNode arrayToDynArrays -n atdaId`;
                }

                setAttr -type "string" ($arrayNodePos +".name") "position";
                setAttr -type "string" ($arrayNodeRot +".name") "rotation";
                setAttr -type "string" ($arrayNodeSca +".name") "scale";
                if (`checkBoxGrp -q -v1 boxVisId`)
                {
                setAttr -type "string" ($arrayNodeVis +".name") "visibility";
                setAttr -type "string" ($arrayNodeId +".name") "objectIndex";
                }
                
                connectAttr -force ($nodeName +".inRotationPP") ($arrayNodeRot +".inArray");
                connectAttr -force ($nodeName +".positions") ($arrayNodePos +".inArray");
                connectAttr -force ($nodeName +".inScalePP") ($arrayNodeSca +".inArray");
                if (`checkBoxGrp -q -v1 boxVisId`)
                {
                connectAttr -force ($nodeName +".inVisibilityPP") ($arrayNodeVis +".inArray");
                connectAttr -force ($nodeName +".inIdPP") ($arrayNodeId +".inArray");
                }
                
                $instancer = `createNode instancer`;
                
                if (`checkBoxGrp -q -v1 boxVisId`)
                {
                connectAttr -force ($arrayNodeId +".outDynamicArrays") ($arrayNodeVis +".inDynamicArrays");
                connectAttr -force ($arrayNodeVis +".outDynamicArrays") ($arrayNodeSca +".inDynamicArrays");
                }
                connectAttr -force ($arrayNodeSca +".outDynamicArrays") ($arrayNodeRot +".inDynamicArrays");
                connectAttr -force ($arrayNodeRot +".outDynamicArrays") ($arrayNodePos +".inDynamicArrays");
                connectAttr -force ($arrayNodePos +".outDynamicArrays") ($instancer +".inputPoints");
                
                setAttr ($nodeName +".outputPPMode") 1;
        
                select $nodeName;
                string $shapesForInstancer[] = `ls -sl`;
                for ($obj in $shapesForInstancer) {
                    instancer -e -a -obj $obj $instancer;
                }
                select $instancer;
                }

            else
                {
                warning "You need to install SOuP to use this feature.";
                }
        }
        
        
        if ($whichCMD == 3)
        {
        
        string $waiterConnection;
        string $connectionType;
        int $chainType = `getAttr ($nodeName+".chainType")`;
        if ($chainType == 1) {
        $waiterConnection = ".inputArray";
        $connectionType = "Pos";
        }
         
        if ($chainType == 2) {
        $waiterConnection = ".inRotationPP";
        $connectionType = "Rot";
        }
        
        if ($chainType == 3) {
        $waiterConnection = ".inScalePP";
        $connectionType = "Scale";
        }
        
        if ($chainType == 5) {
        $waiterConnection = ".inVisibilityPP";
        $connectionType = "";
        }
        
        if ($chainType == 4) {
        $waiterConnection = ".inIdPP";
        $connectionType = "ID";
        }
        
        string $selectedNode = `optionMenu -q -value nodeTypeMenu`;
        string $newNode;
        if ($selectedNode == "Noise") {
            string $nodeType = ("MASH_Noise_"+$connectionType);
            $newNode = `createNode MASH_Noise -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Distribute") {
            string $nodeType = ("MASH_Distribute_"+$connectionType);
            $newNode = `createNode MASH_Distribute -n $nodeType`;
            setAttr  ($newNode+".biasRamp[1].biasRamp_Position") 1;
            setAttr  ($newNode+".biasRamp[1].biasRamp_FloatValue") 1;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Trigonometry") {
            string $nodeType = ("MASH_Trig_"+$connectionType);
            $newNode = `createNode MASH_Trig -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Curve") {
            string $nodeType = ("MASH_Curve_"+$connectionType);
            $newNode = `createNode MASH_Curve -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "ID") {
            string $nodeType = ("MASH_ID_"+$connectionType);
            $newNode = `createNode MASH_ID -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Inheritance") {
            string $nodeType = ("MASH_Inherit_"+$connectionType);
            $newNode = `createNode MASH_Inherit -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Delay") {
            string $nodeType = ("MASH_Delay_"+$connectionType);
            $newNode = `createNode MASH_Delay -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Orient") {
            string $nodeType = ("MASH_Orient_"+$connectionType);
            $newNode = `createNode MASH_Orient -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Spring") {
            string $nodeType = ("MASH_Spring_"+$connectionType);
            $newNode = `createNode MASH_Spring -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Mute") {
            string $nodeType = ("MASH_Mute_"+$connectionType);
            $newNode = `createNode MASH_Mute -n $nodeType`;
        }
        if ($selectedNode == "Random") {
            string $nodeType = ("MASH_Random_"+$connectionType);
            $newNode = `createNode MASH_Random -n $nodeType`;
            connectAttr -f time1.outTime ($newNode +".time");
        }
        if ($selectedNode == "Visibility") {
            string $nodeType = ("MASH_Visibility_"+$connectionType);
            $newNode = `createNode MASH_Visibility -n $nodeType`;
        }
        
        string $incoming[] = `listConnections -d 0 -s 1 ($nodeName+$waiterConnection)`;
        
        if (size($incoming) > 0) {
            if ($connectionType == "Rot"){
                if ($selectedNode == "Spring") {
                connectAttr -force ($incoming[0] +".outRotationPP") ($newNode +".inputArray");
                setAttr ($newNode +".damping") 0.85;
                setAttr ($newNode +".stiffness") 0.05;
                }
                else {
                connectAttr -force ($incoming[0] +".outputArray") ($newNode +".inputArray");
                }
            }
            else {
            connectAttr -force ($incoming[0] +".outputArray") ($newNode +".inputArray");
            }
        }
        
        if ($selectedNode == "Orient") {
            string $incomingOr[] = `listConnections -d 0 -s 1 ($nodeName+".inputArray")`;
            /*
            if (size($incomingOr) > 0) {
                print "connecting incoming rotation";
                connectAttr -force ($incomingOr[0] +".outputArray") ($newNode +".inputArray");
            }*/
            connectAttr -force ($newNode +".outRotationPP") ($nodeName +".inRotationPP");
            string $conns[] = `listConnections -d 1 ($nodeName +".setMessage")`;     
            int $numConns = `size($conns)`;
            int $nextCon = $numConns+1;
            connectAttr -force ($newNode +".message") ($nodeName +".setMessage[" + $nextCon + "]");
            }
        else if ($selectedNode == "Curve") {
            connectAttr -force ($newNode +".outputArray") ($nodeName +".inputArray");
            connectAttr -force ($newNode +".outRotationPP") ($nodeName +".inRotationPP");
            string $conns[] = `listConnections -d 1 ($nodeName +".setMessage")`;    
            int $numConns = `size($conns)`;
            int $nextCon = $numConns+1;
            connectAttr -force ($newNode +".message") ($nodeName +".setMessage[" + $nextCon + "]");
        }
        else if ($selectedNode == "Visibility") {
            connectAttr -force ($newNode +".outVisibilityPP") ($nodeName +".inVisibilityPP");
            string $conns[] = `listConnections -d 1 ($nodeName +".setMessage")`;
            int $numConns = `size($conns)`;
            int $nextCon = $numConns+1;
            connectAttr -force ($newNode +".message") ($nodeName +".setMessage[" + $nextCon + "]");
        }
        else if ($selectedNode == "ID") {
            connectAttr -force ($newNode +".outputArray") ($nodeName +".inIdPP");
            string $conns[] = `listConnections -d 1 ($nodeName +".setMessage")`;
            int $numConns = `size($conns)`;
            int $nextCon = $numConns+1;
            connectAttr -force ($newNode +".message") ($nodeName +".setMessage[" + $nextCon + "]");
        }
        else {
            connectAttr -force ($newNode +".outputArray") ($nodeName + $waiterConnection);
            string $conns[] = `listConnections -d 1 ($nodeName +".setMessage")`;    
            int $numConns = `size($conns)`;
            int $nextCon = $numConns+1;
            connectAttr -force ($newNode +".message") ($nodeName +".setMessage[" + $nextCon + "]");
        }
        evalDeferred("dgdirty "+$newNode);
        }
        
        }
        
        global proc AEwaiterDummyConnections (string $nodeName)
        {
        string $nodeNameTrunc[];
        tokenize($nodeName, ".", $nodeNameTrunc);
        
        string $connections[] = `listConnections -d on -s on ($nodeName +"positions")`;
        if (size($connections) == 0)
        {
       
        string $pointNode = `createNode MASH_Points`;
        connectAttr -force ($nodeName +"positions") ($pointNode +".positions");
        
        
        }
        
        }
        
        global proc AEwaiterDummyConnectionsEdit (string $nodeName)
        {
        
        string $nodeNameTrunc[];
        tokenize($nodeName, ".", $nodeNameTrunc);
        
        string $connections[] = `listConnections -d on -s on ($nodeName +"positions")`;
        if (size($connections) == 0)
        {
        
        
        string $pointNode = `createNode MASH_Points`;
        connectAttr -force ($nodeName +"positions") ($pointNode +".positions");
        
        
        }
        }
        
        //used globally by the plugins to creat input fields.
        global proc createInputs_MASH(string $attr, string $attrString)
        {
        
        rowColumnLayout -nc 3 -cw 1 145 -cw 2 210 -cw 3 25 layoutMASHinputs;
		string $attrLabel[];
		tokenize $attrString "." $attrLabel;
		string $destConnection[] = `listConnections -s 1 -d 0 -p 1 $attr`;
        
		if(size($destConnection) > 0)
		{
        text -l ($attrLabel[1]+" ") -al "right";
        textField -tx $destConnection[0] -ed 0;
        symbolButton -image "inArrow.xpm" -c ("select "+$destConnection[0]);
		}
		else
		{
        text -l ($attrLabel[1]+" ") -al "right";
        textField -tx "" -ed 0;
        symbolButton -image "inArrow.xpm" -en 0;
		}
        
        setParent ..;
        }
        
        global proc editInputs_MASH(string $attr, string $attrString)
        {
        deleteUI -layout layoutMASHinputs;
        createInputs_MASH($attr, $attrString);
        //evalDeferred("refreshAE");
        }
        
        ''')
    
    def __init__( self ):
   
        omMPx.MPxNode.__init__( self )
    
    def compute(self, plug, dataBlock):
        
        
        #if plug != mashWaiter.positions:
        #    return om.kUnknownParameter
        
        whichTransform = dataBlock.inputValue(mashWaiter.aNodeEffects).asInt()
        numOutputsValue = dataBlock.inputValue(mashWaiter.numOutputs).asInt()

        #whichChainValue = dataBlock.inputValue(mashDistribute.aAddToChain).asInt()
        outputPPFlag = dataBlock.inputValue(mashWaiter.outputPPMode).asBool()
        
        # OUTPUTS
        
        #create  and set the output position array
        outResultArray = om.MVectorArray()
        outResultArray.clear()
        outDData = dataBlock.outputValue(mashWaiter.positions)
        nData = om.MFnVectorArrayData()
        
        #create  and set the output rotation array
        rotResultArray = om.MVectorArray()
        rotResultArray.clear()
        rotDData = dataBlock.outputValue(mashWaiter.outRotationPP)
        rotData = om.MFnVectorArrayData()
        
        #create  and set the output scale array
        scaleResultArray = om.MVectorArray()
        scaleResultArray.clear()
        scaleDData = dataBlock.outputValue(mashWaiter.outScalePP)
        scaleData = om.MFnVectorArrayData()
        
        #create  and set the output visibility array
        visResultArray = om.MDoubleArray()
        visResultArray.clear()
        visDData = dataBlock.outputValue(mashWaiter.outVisibilityPP)
        visData = om.MFnDoubleArrayData()
        
        #create  and set the output id array
        idResultArray = om.MVectorArray()
        idResultArray.clear()
        idDData = dataBlock.outputValue(mashWaiter.outIdPP)
        idData = om.MFnVectorArrayData()
        
        #create  and set the output ids array
        idsResultArray = om.MIntArray()
        idsResultArray.clear()
        idsDData = dataBlock.outputValue(mashWaiter.ids)
        idsData = om.MFnIntArrayData()
        
        #INPUTS
        #Get input array
        inDArrData = dataBlock.inputValue(self.inputArray).data()
        inDArrFn = om.MFnVectorArrayData(inDArrData)
        inDArray = inDArrFn.array()
        
        #Get rotaion array
        rotDArrData = dataBlock.inputValue(self.inRotationPP).data()
        rotDArrFn = om.MFnVectorArrayData(rotDArrData)
        rotDArray = rotDArrFn.array()
        
        #Get visibility array
        visDArrData = dataBlock.inputValue(self.inVisibilityPP).data()
        visDArrFn = om.MFnDoubleArrayData(visDArrData)
        visDArray = visDArrFn.array()
        
        #Get id array
        idDArrData = dataBlock.inputValue(self.inIdPP).data()
        idDArrFn = om.MFnVectorArrayData(idDArrData)
        idDArray = idDArrFn.array()
        
        #Get scale array
        scaDArrData = dataBlock.inputValue(self.inScalePP).data()
        scaDArrFn = om.MFnVectorArrayData(scaDArrData)
        scaDArray = scaDArrFn.array()
        
        #get the initial position array
        initialDArrData = dataBlock.inputValue(self.initialState).data()
        initalDArrFn = om.MFnVectorArrayData(initialDArrData)
        initialDArray = initalDArrFn.array()

        thisNode = self.thisMObject()
        selList = om.MSelectionList()
        fnSet = om.MFnSet(thisNode)
        setMembers = fnSet.getMembers(selList,0)
        iter = om.MItSelectionList(selList)
        picked = om.MObjectArray()
        dependNode = om.MObject()
        
        while not iter.isDone():
            # Get the selected dependency node
            iter.getDependNode(dependNode)
            #save it for later
            picked.append(dependNode)
            iter.next()

        numberOfOutputs = selList.length()

        #echo compatability requires some additional counters to reuse visibility and rotational information as it is not echoed.
        v = 0
        r = 0
        s = 0
        id = 0
    
        thisNode = self.thisMObject()
        plugMessage = om.MPlug(thisNode,self.setMessage)
        numberOfconnections = plugMessage.numConnectedElements()
        
        
        for i in range(numberOfconnections):
            inputs = om.MPlugArray()
            inputs.clear()
            source = plugMessage.connectionByPhysicalIndex(i)
            source.connectedTo( inputs, True, False )
            sourceNode = inputs[0].node()
            setDepNode = om.MFnDependencyNode(sourceNode)
            setName = setDepNode.name()
            mc.setAttr( (setName+".inIterations"), numOutputsValue )
        
        
        
        if (outputPPFlag!=1):
            mc.setAttr( (self.name()+".numberOfOutputs"), numberOfOutputs )
                    
        #Dish out the data if there is any...
        if (inDArray.length() > 0):
      
            for i in range(inDArray.length()):
                #add ID for metaballs
                idsResultArray.append(i)
                
                #read in the data from the array
                testRead = om.MVector(0.0,0.0,0.0)
                testRead = inDArray[i]
                
                resultX = testRead.x
                resultY = testRead.y
                resultZ = testRead.z
                
                if (initialDArray.length() > 0):
                    testRead = om.MVector(0.0,0.0,0.0)
                    testRead = initialDArray[i]
                    resultX +=  testRead.x 
                    resultY +=  testRead.y 
                    resultZ +=  testRead.z
                
                #set the outputs
                if (outputPPFlag!=1):
                    dependNode = picked[i]
                    fnDependNode = om.MFnDependencyNode(dependNode)
                    
                    fnDependNode.findPlug("tx").setFloat(resultX)
                    fnDependNode.findPlug("ty").setFloat(resultY)
                    fnDependNode.findPlug("tz").setFloat(resultZ)
                
                thisResult = om.MVector(resultX,resultY,resultZ)
            
                #add the result to the out array
                outResultArray.append(thisResult)

        #set the rotation values (in radians!) if something is connected to the inRotationPP attribute
        if (rotDArray.length() > 0):
            for i in range(rotDArray.length()):
                #print "rot array calcing"
                rotRead = om.MVector(0.0,0.0,0.0)
                
                rotRead = rotDArray[r]

                if (outputPPFlag!=1):
                    rotX = om.MAngle(rotRead.x, om.MAngle.kDegrees).asRadians()
                    rotY = om.MAngle(rotRead.y, om.MAngle.kDegrees).asRadians()
                    rotZ = om.MAngle(rotRead.z, om.MAngle.kDegrees).asRadians()
                    dependNode = picked[r]
                    fnDependNode = om.MFnDependencyNode(dependNode)
                    fnDependNode.findPlug("rx").setFloat(rotX)
                    fnDependNode.findPlug("ry").setFloat(rotY)
                    fnDependNode.findPlug("rz").setFloat(rotZ)
            
                #if i < 1:
                #    print "rot x is ", rotRead.x
                    
                rotResultArray.append(rotRead)
                
                r += 1
                if r == rotDArray.length():
                    r = 0
    
        #set the visibility values if something is connected to the inVisibilityPP attribute
        if (visDArray.length() > 0):
            for i in range(visDArray.length()):
                visRead = 0
                
                visRead = visDArray[v]

                result = int(visRead)
                if (outputPPFlag!=1):
                    dependNode = picked[v]
                    fnDependNode = om.MFnDependencyNode(dependNode)
                    fnDependNode.findPlug("visibility").setInt(result)
                
                visResultArray.append(result)
                
                v += 1
                if v == visDArray.length():
                    v = 0
    
        #set the scale values if something is connected to the inScalePP attribute
        if (scaDArray.length() > 0):
            for i in range(scaDArray.length()):
                scaRead = om.MVector(0.0,0.0,0.0)
                
                scaRead = scaDArray[s]

                if (outputPPFlag!=1):
                    dependNode = picked[s]
                    fnDependNode = om.MFnDependencyNode(dependNode)
                    fnDependNode.findPlug("sx").setFloat(scaRead.x)
                    fnDependNode.findPlug("sy").setFloat(scaRead.y)
                    fnDependNode.findPlug("sz").setFloat(scaRead.z)
            
                scaleResultArray.append(scaRead)
                
                s += 1
                if s == scaDArray.length():
                    s = 0

                
        #set the id values if something is connected to the inIdPP attribute
        if (idDArray.length() > 0):
            for i in range(idDArray.length()):
                idRead = 0
                
                idRead = idDArray[id]
                
                result = int(idRead.x)
                
                idResultArray.append(idRead)
                
                id += 1
                if id == idDArray.length():
                    id = 0
                    



        rotDataObj = rotData.create(rotResultArray)
        rotDData.setMObject(rotDataObj)
                
        #if plug == mashWaiter.positions:
        nDataObj = nData.create(outResultArray)
        outDData.setMObject(nDataObj)
            
        scaleDataObj = scaleData.create(scaleResultArray)
        scaleDData.setMObject(scaleDataObj)
        
        idDataObj = idData.create(idResultArray)
        idDData.setMObject(idDataObj)
    
        visDataObj = visData.create(visResultArray)
        visDData.setMObject(visDataObj)
    
        idsDataObj = idsData.create(idsResultArray)
        idsDData.setMObject(idsDataObj)
        
        #have a wash
        dataBlock.setClean(plug)

def nodeCreator():

    return omMPx.asMPxPtr( mashWaiter() )

def nodeInitializer():
    
    
    defaultVectorArray = om.MVectorArray();
    dArrayDataFn = om.MFnVectorArrayData();
    dArrayDataFn.create( defaultVectorArray );
    
    defaultDoubleArray = om.MDoubleArray();
    dDoubleArrayDataFn = om.MFnDoubleArrayData();
    dDoubleArrayDataFn.create( defaultDoubleArray );
    
    defaultIntArray = om.MIntArray();
    dIntArrayDataFn = om.MFnIntArrayData();
    dIntArrayDataFn.create( defaultIntArray );
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.inputArray = tAttr.create("inputArray", "inArray", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.inputArray)
    
    nAttr = om.MFnNumericAttribute()
    mashWaiter.output = nAttr.create( "dummyOutput", "out", om.MFnNumericData.k3Float, 0 )
    nAttr.setStorable(1)
    nAttr.setWritable(1)
    nAttr.setHidden(1)
    mashWaiter.addAttribute( mashWaiter.output )
       
    nAttr = om.MFnNumericAttribute()
    mashWaiter.numOutputs = nAttr.create("numberOfOutputs", "numO", om.MFnNumericData.kInt, 0)
    nAttr.setStorable(1)
    nAttr.setHidden(0)
    nAttr.setKeyable(1)
    nAttr.setWritable(1)
    mashWaiter.addAttribute( mashWaiter.numOutputs )
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.positions = tAttr.create("positions", "ppA", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.positions)

    tAttr = om.MFnTypedAttribute()
    mashWaiter.inRotationPP = tAttr.create("inRotationPP", "inRotPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.inRotationPP)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.inVisibilityPP = tAttr.create("inVisibilityPP", "inVisPP", om.MFnData.kDoubleArray, dDoubleArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.inVisibilityPP)
    
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.inScalePP = tAttr.create("inScalePP", "inScPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.inScalePP)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.inIdPP = tAttr.create("inIdPP", "inIdPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setWritable(1)
    tAttr.setStorable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.inIdPP)
    
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.outRotationPP = tAttr.create("outRotationPP", "outRotPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.outRotationPP)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.outVisibilityPP = tAttr.create("outVisibilityPP", "outVisPP", om.MFnData.kDoubleArray, dDoubleArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.outVisibilityPP)
    
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.outScalePP = tAttr.create("outScalePP", "outScaPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.outScalePP)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.outIdPP = tAttr.create("outIdPP", "outIdPP", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.outIdPP)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.ids = tAttr.create("ids", "id", om.MFnData.kIntArray, dIntArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(0)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.ids)
    
    tAttr = om.MFnTypedAttribute()
    mashWaiter.initialState = tAttr.create("initialState", "initSt", om.MFnData.kVectorArray, dArrayDataFn.object())
    tAttr.setStorable(1)
    tAttr.setWritable(1)
    tAttr.setReadable(1)
    tAttr.setKeyable(0)
    tAttr.setHidden(0)
    mashWaiter.addAttribute( mashWaiter.initialState)
    
    mAttr = om.MFnMessageAttribute()
    mashWaiter.setMessage = mAttr.create( "setMessage", "smsg" )
    mAttr.setArray(1)
    mAttr.setConnectable(1)
    mAttr.setStorable(1)
    mAttr.setReadable(1)
    mAttr.setWritable(1)
    mashWaiter.addAttribute( mashWaiter.setMessage )
    
    nAttr = om.MFnNumericAttribute()
    mashWaiter.outputPPMode = nAttr.create ( "outputPPMode", "outPPM", om.MFnNumericData.kBoolean, 0 )
    mashWaiter.addAttribute ( mashWaiter.outputPPMode )
    
    nAttr = om.MFnNumericAttribute()
    mashWaiter.aNodeEffects = nAttr.create("nodeEffects", "noEff", om.MFnNumericData.kInt, 0)
    mashWaiter.addAttribute( mashWaiter.aNodeEffects )
    mashWaiter.attributeAffects( mashWaiter.aNodeEffects, mashWaiter.output)
    
    typedAttr = om.MFnTypedAttribute()
    enumAttr = om.MFnEnumAttribute()
    mashWaiter.aAddToChain = enumAttr.create("chainType", "cht", om.MFnData.kNumeric)
    enumAttr.addField('Position', 1)
    enumAttr.addField('Rotation', 2)
    enumAttr.addField('Scale', 3)
    enumAttr.addField('Instancer ID', 4)
    #enumAttr.addField('Visibility', 5)
    mashWaiter.randomeData = typedAttr.create("chainTypeData", "ctd", om.MFnData.kString)
    mashWaiter.addAttribute ( mashWaiter.aAddToChain )
    mashWaiter.attributeAffects( mashWaiter.aAddToChain, mashWaiter.positions)
    

    
    mashWaiter.attributeAffects( mashWaiter.inScalePP, mashWaiter.positions)
    mashWaiter.attributeAffects( mashWaiter.inVisibilityPP, mashWaiter.positions)
    mashWaiter.attributeAffects( mashWaiter.inRotationPP, mashWaiter.positions)
    mashWaiter.attributeAffects( mashWaiter.inputArray, mashWaiter.positions) 
    mashWaiter.attributeAffects( mashWaiter.inputArray, mashWaiter.positions)

def initializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject, "Ian_Waters", "1.0", "Any")
    
    try:
        #The cool bit here is setting the kNodeType at the end
        mplugin.registerNode( kPluginNodeTypeName, mashWaiterId, nodeCreator, nodeInitializer, omMPx.MPxNode.kObjectSet )
    
    except:
        sys.stderr.write( "Failed to register node: %s" % kPluginNodeTypeName )
        raise

def uninitializePlugin(mobject):
    
    mplugin = omMPx.MFnPlugin(mobject)
    
    try:
        mplugin.deregisterNode( mashWaiterId )
    
    except:
        sys.stderr.write( "Failed to deregister node: %s" % PluginNodeTypeName )
        
        
        

    